import datetime
import os
import re
import struct
from typing import Any

import olefile as ole
import pandas as pd
import xmltodict

from allotropy.named_file_contents import NamedFileContents

# Patterns to extract cycle number, window number, and curve number
cycle_pattern = re.compile(r"_Cycle (\d+)")
window_pattern = re.compile(r"_Window (\d+)")
curve_pattern = re.compile(r"_Curve (\d+)")


def get_cycle_data(
    cycles: list[dict[str, Any]],
    cycle_number: int,
    curve_number: int,
    window_number: int,
) -> Any:
    """
    Gets the cycle, window and curve number
    :param cycles: cycle list
    :param cycle_number: cycle number
    :param curve_number: curve number
    :param window_number: window number
    :return: cycle dict
    """
    for cycle in cycles:
        if (
            cycle["cycle_number"] == cycle_number
            and cycle["cycle_number"] == curve_number
            and cycle["window_number"] == window_number
        ):
            return cycle

    new_cycle = {
        "cycle_number": cycle_number,
        "curve_number": curve_number,
        "window_number": curve_number,
    }
    cycles.append(new_cycle)

    return new_cycle


def get_sample_data(sample_dict: dict[str, Any]) -> list[dict[str, Any]]:
    """
    Gets Sample data from the dictionary
    :param sample_data: dictionary containing sample data
    :return: structured dictionary containing sample data
    """
    sample_data = []
    sample_details = []
    positions = sample_dict["positions"]
    position_order = sample_dict["position_order"]

    for key, value in positions.items():
        samples = {}
        if key.endswith("sample_pos"):
            samples["rack"] = value
            cycle = match.group() if (match := re.search(r"Cycle\d+", key)) else None

            samples["cycle_number"] = (
                "Cycle" + str(int(cycle.split("Cycle")[1].split("@")[0]) + 1)
                if cycle is not None and "Cycle" in cycle
                else None
            )
            for order in position_order.keys():
                if value in order:
                    if "running buffer" and "Startup" in order:
                        samples["role"] = "blank role"
                    elif "Sample" in order:
                        samples["sample_name"] = order.split("\n")[0]
                        samples["role"] = "sample role"

            sample_details.append(samples)

    sorted_data = sorted(
        sample_details, key=lambda d: int(d["cycle_number"].replace("Cycle", ""))
    )
    sample_iter = iter(sample_dict["sample_data"])
    for item in sorted_data:
        if item["role"] == "blank role":
            sample_data.append(item)
        else:
            sample = next(sample_iter, None)
            if sample:
                new_item = item.copy()
                if sample["Sample id"] == item["sample_name"]:
                    new_item["sample_name"] = sample.get("Sample id", None)
                    new_item["concentration"] = (
                        sample["Conc."].replace(",", ".") if sample["Conc."] else None
                    )
                    new_item["molecular_weight"] = sample.get("MW", None)
                    sample_data.append(new_item)

    return sample_data


def process_and_rearrange_xml_data(data: dict[str, Any]) -> dict[str, Any]:
    """
    Processes and Rearranges xml data to form structured dictionary
    :param data: The data containing xml details
    :return: structured dictionary
    """
    # Initialize the output dictionary
    output = {}

    # Process the input dictionary based on the provided code structure
    for data_type in ["string", "integer", "boolean"]:
        if data_type in data:
            # Handle cases where the data type is a list of dictionaries
            if isinstance(data[data_type], list):
                for entry in data[data_type]:
                    # Convert @key and @value pairs, handle boolean conversion
                    if "@key" in entry and "@value" in entry:
                        if data_type == "boolean":
                            # Convert 'true'/'false' to actual boolean values
                            output[entry["@key"]] = entry["@value"].lower() == "true"
                        elif data_type == "integer":
                            output[entry["@key"]] = int(entry["@value"])
                        else:
                            output[entry["@key"]] = entry["@value"]
                    else:
                        output.update(entry)
            # Handle single dictionary entries
            elif isinstance(data[data_type], dict):
                if "@key" in data[data_type] and "@value" in data[data_type]:
                    if data_type == "boolean":
                        output[data[data_type]["@key"]] = (
                            data[data_type]["@value"].lower() == "true"
                        )
                    elif data_type == "integer":
                        output[data[data_type]["@key"]] = int(data[data_type]["@value"])
                    else:
                        output[data[data_type]["@key"]] = data[data_type]["@value"]
                else:
                    output.update(data[data_type])

    return output


def get_detection_data(
    detection_settings: dict[str, list[dict[str, Any]]]
) -> dict[str, Any]:
    """
    The function gets detection data
    :param detection_settings: dictionary containing detection data
    :return: structured dictionary containing detection data
    """
    detection_data = {}
    # Iterate through each item in 'dataItem'
    for item in detection_settings["dataItem"]:
        # Check if the 'value' in 'valueItem' is not None
        if item["valueItem"]["value"] is not None:
            # Add the '@id' and 'value' to the output dictionary
            detection_data[item["@id"]] = item["valueItem"]["value"]

    return detection_data


def get_data_item(data_item: dict[str, Any]) -> dict[str, Any]:
    """
    The function gets the data item
    :param data_item: dictionary containing data item
    :return: structured dictionary of data item
    """
    data_time = {}
    for item in data_item["dataItem"]:
        value_data = item["valueItem"]["value"]

        if isinstance(value_data, dict):
            value_dict = {"value": value_data.get("#text")}
        else:
            value_dict = {"value": value_data}

        if "min" in item["valueItem"]:
            value_dict["min"] = item["valueItem"]["min"]
        if "max" in item["valueItem"]:
            value_dict["max"] = item["valueItem"]["max"]

        data_time[item["@id"]] = value_dict
    return data_time


def decode_application_data(
    application_template_data: dict[str, Any]
) -> tuple[dict[str, dict[str, Any]], list[dict[str, Any]]]:
    """
    decodes app data template application stream data
    :param application_template_data: string app data template application stream data
    :return: dict app data template application stream data
    """
    total_samples: list[dict[str, Any]] = []
    application_template = {}
    sample_data = {}
    for application_value in application_template_data.values():

        if "HtmlPreview" in application_value["Properties"]:
            if application_value["Properties"]["TypeName"] == "Kinetics/Affinity":

                sample_data["run_order"] = (
                    application_value["Properties"]["HtmlPreview"]
                    .split("Run order")[1]
                    .split("\xa0\n\xa0")[0]
                    .split("\n\n")[0]
                    .split("\xa0")[1]
                    .strip()
                )
                lines = (
                    application_value["Properties"]["HtmlPreview"]
                    .split("Run order")[1]
                    .split("\xa0\n\xa0")[0]
                    .split("\n\n")[1]
                    .split("\n")
                )
                header = re.split(r"\s{2,}", lines[0].strip())
                # Extract the rows
                rows = [re.split(r"\s{2,}", line.strip()) for line in lines[1:]]
                rows = [row[1:] for row in rows]
                # Convert to a list of dictionaries
                sample_details = [dict(zip(header, row, strict=True)) for row in rows]
                sample_data["sample_data"] = sample_details

            del application_value["Properties"]["HtmlPreview"]

        if "xmlBag" in application_value:
            for xml_name in application_value["xmlBag"]:
                if xml_name["@name"] == "_Racks":
                    application_template["racks"] = process_and_rearrange_xml_data(
                        xml_name
                    )
                elif xml_name["@name"] == "_Positions":
                    sample_data["positions"] = process_and_rearrange_xml_data(xml_name)
                elif xml_name["@name"] == "_PositionOrder":
                    sample_data["position_order"] = process_and_rearrange_xml_data(
                        xml_name
                    )
                elif "Flowcell" in xml_name["@name"]:
                    flow_data = process_and_rearrange_xml_data(xml_name)
                    if flow_data["UseFlowcell"]:
                        application_template[xml_name["@name"]] = flow_data
                elif xml_name["@name"] == "MethodRun":
                    for xml_string in xml_name["string"]:
                        xml_string["@value"] = xmltodict.parse(xml_string["@value"])
                        if "methodSettings" in xml_string["@value"]["method"]:
                            application_template["detection"] = get_detection_data(
                                xml_string["@value"]["method"]["methodSettings"][
                                    "detectionSettings"
                                ]
                            )
                            measurement_settings = get_data_item(
                                xml_string["@value"]["method"]["methodSettings"][
                                    "dataItems"
                                ]
                            )
                            application_template[
                                "RackTemperature"
                            ] = measurement_settings["RackTemperature"]
                            application_template["BaselineFlow"] = measurement_settings[
                                "BaselineFlow"
                            ]
                            application_template[
                                "DataCollectionRate"
                            ] = measurement_settings["DataCollectionRate"]
                            application_template[
                                "MoleculeWeightUnit"
                            ] = measurement_settings["MoleculeWeightUnit"]
                elif xml_name["@name"] == "_SystemPreparations":
                    application_template[
                        "system_preparations"
                    ] = process_and_rearrange_xml_data(xml_name)
                elif xml_name["@name"] == "_PrepareRun":
                    application_template["prepare_run"] = {
                        key: value
                        for key, value in process_and_rearrange_xml_data(
                            xml_name
                        ).items()
                        if "Buffer" in key and value != -1
                    }

        application_template["properties"] = application_value["Properties"]
        if (
            application_value["Properties"]["TypeName"] == "Kinetics/Affinity"
            and sample_data["run_order"] == "As entered"
        ):
            total_samples = get_sample_data(sample_data)
    return application_template, total_samples


def convert_datetime(days_str: str) -> str:
    """
    converts the str to iso datetime format
    :param days_str: string of datetime
    :return: datetime
    """
    days = float(days_str)
    start_date = datetime.datetime(
        year=1899, month=12, day=30, tzinfo=datetime.timezone.utc
    )
    delta = datetime.timedelta(days=days)
    result_date = start_date + delta
    iso_date_time = result_date.isoformat()
    return iso_date_time


def extract_stream_data(data: str) -> dict[str, Any]:
    """
    Decodes various types of stream data.
    :param data: string stream data
    :return: dict with decoded data
    """
    data_dict = {}
    for line in data.strip().split("\n"):
        key, value = line.split("=")
        data_dict[key] = value
        if any(
            time_key in key.lower()
            for time_key in (
                "time",
                "date",
            )
        ) and isinstance(value, str):

            data_dict[key] = convert_datetime(value)

    return data_dict


def get_dip_data(input_data: str) -> dict[str, Any]:
    """
    Parses the input string data and returns a structured JSON of the decoded data.
    :param input_data: string stream data
    :return: dict with structured JSON data
    """
    lines = input_data.split("\n")
    result: dict[str, Any] = {}
    norm_data: list[dict[str, Any]] = []
    raw_data: list[dict[str, Any]] = []
    current_section = None
    max_length = 2
    for line in lines:
        each_line = line.strip()
        if each_line.startswith("Count"):
            result["count"] = int(each_line.split("=")[1])
        elif each_line.startswith("Timestamp1"):
            timestamp_value = each_line.split("=")[1]
            result["timestamp"] = convert_datetime(timestamp_value)
        elif "Norm" in each_line:
            current_section = "norm"
        elif "Raw" in each_line:
            current_section = "raw"
        else:
            parts = each_line.split(";")
            if (
                len(parts) == max_length
                and len(index_part := parts[0].split("=")) == max_length
            ):
                sweep_row, flow_cell = index_part[1].split(",")
                response = list(map(int, parts[1].split(",")))
                data = {
                    "flow_cell": flow_cell,
                    "sweep_row": sweep_row,
                    "response": response,
                }
                (norm_data if current_section == "norm" else raw_data).append(data)

    result.update({"norm_data": norm_data, "raw_data": raw_data})
    return result


def get_r_point_data(r_point_data: str) -> list[dict[str, Any]]:
    """
    decodes the r-point data
    :param r_point_data: str of r point data
    :return: list of r_point dictionary
    """
    r_data_list = []
    lines = r_point_data.strip().split("\n")
    header = lines[0].split("\t")

    for line in lines[1:]:
        values = line.split("\t")
        data_dict = dict(zip(header, values, strict=True))

        r_data_list.append(data_dict)

    return r_data_list


def decode_data(named_file_contents: NamedFileContents) -> dict[str, Any]:
    """
    Decodes the proprietary file into a structured dict
    :param named_file_contents: The named file contents containing the input file details
    :return: structured dictionary of decoded data
    """
    file_path = named_file_contents.original_file_path
    base_name = os.path.basename(file_path).split(".blr")[0]
    dir_name = os.path.dirname(file_path)
    intermediate_json: dict[str, Any] = {}
    content = ole.OleFileIO(file_path)
    streams = content.listdir()
    cycles: list[dict[str, Any]] = []
    cycle_df_list = []
    cycle_details = []
    for stream in streams:
        stream_content = content.openstream(stream).read()
        if stream == ["Environment"]:
            intermediate_json["system_information"] = extract_stream_data(
                stream_content.decode("utf-8")
            )
        elif stream == ["Chip"]:
            intermediate_json["chip"] = extract_stream_data(
                stream_content.decode("utf-8")
            )

        elif stream == ["AppData", "Dip"]:
            intermediate_json["dip"] = get_dip_data(stream_content.decode("utf-8"))
        elif stream == ["AppData", "ApplicationTemplate"]:
            (
                intermediate_json["application_template_details"],
                sample_data,
            ) = decode_application_data(xmltodict.parse(stream_content))
            if sample_data:
                intermediate_json["sample_data"] = sample_data
        elif stream == ["RPoint Table"]:
            r_point_data = get_r_point_data(stream_content.decode("utf-8"))
            r_point_dataframe = pd.DataFrame(r_point_data)
            report_point_data = r_point_dataframe.groupby("Cycle")
            for _name, group in report_point_data:
                cycle_data = {}
                cycle_number = group["Cycle"].iloc[0]
                cycle_name = f"_cycle_{cycle_number}"
                rpoint_parquet_path = os.path.join(
                    dir_name, base_name + cycle_name + "_rpoint.parquet"
                )
                group.to_parquet(rpoint_parquet_path, index=False)
                cycle_data["r-point_path"] = rpoint_parquet_path
                cycle_data["cycle_number"] = cycle_number
                cycle_details.append(cycle_data)
        cycle_match = cycle_pattern.search(stream[0])
        if not cycle_match:
            continue
        cycle_number = int(cycle_match.group(1))

        for part in stream:
            window_match = window_pattern.search(part)

        if any("_Window" in part for part in stream) and any(
            "_Curve" in part for part in stream
        ):
            window_group = 1
            # gets the curve number
            curve_number = next(
                int(curve_match.group(1))
                for part in stream
                if (curve_match := curve_pattern.search(part))
            )
            # gets the window number
            window_number = next(
                int(window_match.group(window_group))
                for part in stream
                if (window_match := window_pattern.search(part))
            )
            # gets the cycle number
            cycle = get_cycle_data(cycles, cycle_number, curve_number, window_number)

            # gets the flow cell
            if "Labels" in str(stream):
                label_list = []
                if len(stream_content) != 0:
                    decoded_stream_content = stream_content.decode("utf-8")
                    for line in decoded_stream_content.strip().split("\n"):
                        label_list.append(line)
                        if "Fc" in line:
                            flow_cell = label_list[0].split("=")[1]

            # gets the xy data as data frame
            elif "XYData" in str(stream):
                xy_data_list = list(
                    struct.unpack("f" * (len(stream_content) // 4), stream_content)
                )
                indexed_xy_data_list = xy_data_list[3:]
                length_of_xydata = len(xy_data_list[3:])
                xy_result_list = indexed_xy_data_list[int(length_of_xydata / 2) :]
                time = indexed_xy_data_list[: int(length_of_xydata / 2)]
                length_of_xy = len(xy_result_list)

                xy_data_frame = pd.DataFrame(
                    {
                        "Flow Cell Number": [flow_cell] * length_of_xy,
                        "Cycle Number": [cycle["cycle_number"]] * length_of_xy,
                        "Curve Number": [cycle["curve_number"]] * length_of_xy,
                        "Window Number": [cycle["window_number"]] * length_of_xy,
                        "Sensorgram (RU)": xy_result_list,
                        "Time (s)": time,
                    }
                )

                cycle_df_list.append(xy_data_frame)

            # gets the segment data
            elif "Segment" in str(stream):
                segment_data_list = list(
                    struct.unpack("f" * (len(stream_content) // 4), stream_content)
                )
                length_of_segment = len(segment_data_list[11:])
                segment_data_frame = pd.DataFrame(
                    {
                        "Flow Cell Number": [flow_cell] * length_of_segment,
                        "Cycle Number": [cycle["cycle_number"]] * length_of_segment,
                        "Curve Number": [cycle["curve_number"]] * length_of_segment,
                        "Window Number": [cycle["window_number"]] * length_of_segment,
                        "Sensorgram (RU)": segment_data_list[11:],
                    }
                )
                cycle_df_list.append(segment_data_frame)
    combined_cycles_df = pd.concat(cycle_df_list, ignore_index=True)

    sensorgram_data = combined_cycles_df.groupby("Cycle Number")
    cycles_data = []
    for _name, group in sensorgram_data:
        if "Time (s)" in group.columns:
            max_new = group["Flow Cell Number"].max()
            last_part = group[group["Flow Cell Number"] == max_new]
            time_map = dict(zip(last_part.index, last_part["Time (s)"], strict=True))
            for new_value in group["Flow Cell Number"].unique():
                if new_value != max_new:
                    indices = group[group["Flow Cell Number"] == new_value].index
                    for i, idx in enumerate(indices):
                        group.at[idx, "Time (s)"] = time_map[
                            list(time_map.keys())[i % len(time_map)]
                        ]
        elif intermediate_json["application_template_details"].get(
            "DataCollectionRate"
        ):
            group["Time (s)"] = group.groupby("Flow Cell Number").cumcount() * (
                1
                / intermediate_json["application_template_details"][
                    "DataCollectionRate"
                ]["value"]
            )
        else:
            group["Time (s)"] = group.groupby("Flow Cell Number").cumcount() + 1
        cycle_number = group["Cycle Number"].iloc[0]
        cycle_name = f"_cycle_{cycle_number}"
        group_data = group[
            [
                "Flow Cell Number",
                "Cycle Number",
                "Curve Number",
                "Window Number",
                "Time (s)",
                "Sensorgram (RU)",
            ]
        ]
        sensorgram_parquet_path = os.path.join(
            dir_name, base_name + cycle_name + "_sensorgram.parquet"
        )
        group_data.to_parquet(sensorgram_parquet_path, index=False)
        cycles_data.append(sensorgram_parquet_path)

    total_cycles = [
        {**d, "sensorgram_path": new_value}
        for d, new_value in zip(cycle_details, cycles_data, strict=True)
    ]
    intermediate_json["cycle_data"] = total_cycles
    intermediate_json["total_cycles"] = cycle_number

    return intermediate_json



from __future__ import annotations

from decimal import Decimal
from pathlib import Path

from allotropy.allotrope.schema_mappers.adm.cell_counting.rec._2024._09.cell_counting import (
    Measurement,
    MeasurementGroup,
    Metadata,
)
from allotropy.parsers.constants import NOT_APPLICABLE
from allotropy.parsers.roche_cedex_hires import constants
from allotropy.parsers.utils.pandas import SeriesData
from allotropy.parsers.utils.uuids import random_uuid_str


def create_metadata(data: SeriesData, file_path: str) -> Metadata:
    path = Path(file_path)
    return Metadata(
        asm_file_identifier=path.with_suffix(".json").name,
        data_system_instance_id=NOT_APPLICABLE,
        file_name=path.name,
        unc_path=file_path,
        device_type=constants.DEVICE_TYPE,
        detection_type=constants.DETECTION_TYPE,
        model_number=constants.MODEL_NUMBER,
        software_name=constants.CEDEX_SOFTWARE,
        product_manufacturer=constants.PRODUCT_MANUFACTURER,
        asset_management_identifier=data[str, "System name"],
        description=data[str, "System description"],
        device_identifier=data[str, "Cedex ID"],
        brand_name=constants.BRAND_NAME,
    )


def create_measurement_groups(data: SeriesData) -> MeasurementGroup:
    # Cell counts are measured in cells/mL, but reported in millions of cells/mL
    viable_cell_density = float(
        Decimal(data[str, "Viable Cell Conc."]) / Decimal("1000000")
    )
    total_cell_density_val = data.get(str, "Total Cell Conc.")
    if total_cell_density_val:
        total_cell_density = float(Decimal(total_cell_density_val) / Decimal("1000000"))
    dead_cell_density_val = data.get(str, "Dead Cell Conc.")
    if dead_cell_density_val:
        dead_cell_density = float(Decimal(dead_cell_density_val) / Decimal("1000000"))

    return MeasurementGroup(
        analyst=data.get(str, "Username"),
        measurements=[
            Measurement(
                measurement_identifier=random_uuid_str(),
                timestamp=data[str, "Date finished"],
                sample_identifier=data[str, "Sample identifier"],
                batch_identifier=data.get(str, "Reactor identifier"),
                group_identifier=data.get(str, "Data set name"),
                viability=data[float, "Viability"],
                viable_cell_count=data.get(float, "Viable Cell Count"),
                viable_cell_density=viable_cell_density,
                total_cell_count=data.get(float, "Total Cell Count"),
                total_cell_density=(
                    total_cell_density if total_cell_density_val else None
                ),
                dead_cell_count=data.get(float, "Dead Cell Count"),
                dead_cell_density=dead_cell_density if dead_cell_density_val else None,
                cell_type_processing_method=data.get(str, "Cell type name"),
                cell_density_dilution_factor=data.get(float, "Dilution"),
                sample_volume_setting=data.get(float, "Sample volume"),
                average_live_cell_diameter=data.get(float, "Avg Diameter"),
                average_compactness=data.get(float, "Avg Compactness"),
                average_area=data.get(float, "Avg Area"),
                average_perimeter=data.get(float, "Avg Perimeter"),
                average_segment_area=data.get(float, "Avg Segm. Area"),
                total_object_count=data.get(float, "Total Object Count"),
                standard_deviation=data.get(float, "Std Dev."),
                aggregate_rate=data.get(float, "Aggregate Rate"),
                sample_draw_time=data.get(str, "Sample draw Time"),
            )
        ],
    )



from dataclasses import dataclass

import pandas as pd

from allotropy.parsers.appbio_quantstudio_designandanalysis.appbio_quantstudio_designandanalysis_reader import (
    DesignQuantstudioReader,
)
from allotropy.parsers.appbio_quantstudio_designandanalysis.structure.generic.structure import (
    WellList,
)


@dataclass(frozen=True)
class RelativeStandardCurveWellList(WellList):
    @classmethod
    def get_well_result_data(cls, reader: DesignQuantstudioReader) -> pd.DataFrame:
        new_data = cls._add_data(
            data=reader.get_non_empty_sheet(cls.get_data_sheet()),
            extra_data=reader.get_non_empty_sheet("Replicate Group Result"),
            columns=[
                "Cq SE",
            ],
        )

        return cls._add_data(
            new_data,
            extra_data=reader.get_non_empty_sheet("RQ Replicate Group Result"),
            columns=[
                "EqCq Mean",
                "Adjusted EqCq Mean",
                "Delta EqCq Mean",
                "Delta EqCq SD",
                "Delta EqCq SE",
                "Delta Delta EqCq",
                "Rq",
                "Rq Min",
                "Rq Max",
            ],
        )



from typing import Any

import pandas as pd

from allotropy.json_to_csv.json_to_csv import json_to_csv
from allotropy.json_to_csv.mapper_config import MapperConfig


def map_json(
    data: dict[str, Any], mapper_config: dict[str, Any]
) -> dict[str, pd.DataFrame | dict[str, Any]]:
    return json_to_csv(data, MapperConfig.create(mapper_config))



from functools import partial

from allotropy.allotrope.models.adm.plate_reader.rec._2024._06.plate_reader import (
    Model,
)
from allotropy.allotrope.schema_mappers.adm.plate_reader.rec._2024._06.plate_reader import (
    Data,
    Mapper,
)
from allotropy.named_file_contents import NamedFileContents
from allotropy.parsers.release_state import ReleaseState
from allotropy.parsers.tecan_magellan.constants import DISPLAY_NAME
from allotropy.parsers.tecan_magellan.tecan_magellan_reader import (
    TecanMagellanReader,
)
from allotropy.parsers.tecan_magellan.tecan_magellan_structure import (
    create_measurement_groups,
    create_metadata,
    MagellanMetadata,
)
from allotropy.parsers.utils.pandas import map_rows
from allotropy.parsers.vendor_parser import VendorParser


class TecanMagellanParser(VendorParser[Data, Model]):
    DISPLAY_NAME = DISPLAY_NAME
    RELEASE_STATE = ReleaseState.RECOMMENDED
    SUPPORTED_EXTENSIONS = TecanMagellanReader.SUPPORTED_EXTENSIONS
    SCHEMA_MAPPER = Mapper

    def create_data(self, named_file_contents: NamedFileContents) -> Data:
        reader = TecanMagellanReader(named_file_contents)
        metadata = MagellanMetadata.create(reader.metadata_lines)
        well_count = len(reader.data)

        return Data(
            create_metadata(metadata, named_file_contents.original_file_path),
            map_rows(
                reader.data,
                partial(
                    create_measurement_groups, metadata=metadata, well_count=well_count
                ),
            ),
        )



import json
from typing import Any

from allotropy.named_file_contents import NamedFileContents
from allotropy.parsers.utils.values import assert_not_none


class BenchlingEmpowerReader:
    SUPPORTED_EXTENSIONS = "json"

    metadata_fields: dict[str, Any]
    injections: list[dict[str, Any]]

    def __init__(self, named_file_contents: NamedFileContents) -> None:
        contents: dict[str, Any] = json.load(named_file_contents.contents)
        values: dict[str, Any] = assert_not_none(contents.get("values"), "values")

        self.metadata_fields = assert_not_none(values.get("fields"), "values/fields")

        # Each injection corresponds to a measurement document
        injections: list[dict[str, Any]] = assert_not_none(
            values.get("injections"), "values/injections"
        )
        for idx, injection in enumerate(injections):
            if "fields" not in injection:
                msg = f"Missing 'fields' for injection at index {idx}"
                raise AssertionError(msg)
            if "InjectionId" not in injection["fields"]:
                msg = f"Missing 'InjectionId' for injection at index {idx}"
                raise AssertionError(msg)
        id_to_injection = {
            injection["fields"]["InjectionId"]: injection["fields"]
            for injection in injections
        }

        # Channels contain fields and chromatograms
        channels: list[dict[str, Any]] = values.get("channels", [])
        for channel in channels:
            fields: dict[str, Any] = assert_not_none(
                channel.get("fields"), "channels/fields"
            )
            if (inj_id := fields.get("InjectionId")) is None:
                msg = f"Expected InjectionId in 'fields' for channel: {fields['ChannelId']}"
                raise AssertionError(msg)
            for key, value in fields.items():
                if (
                    key in id_to_injection[inj_id]
                    and id_to_injection[inj_id][key] != value
                ):
                    msg = f"Mismatch between injection field and channel field for key: {key}"
                    raise AssertionError(msg)
                id_to_injection[inj_id][key] = value
            if "chrom" in channel:
                id_to_injection[inj_id]["chrom"] = channel["chrom"]

        # Results contain peaks and calibration cureves
        results: list[dict[str, Any]] = values.get("results", [])
        for result in results:
            fields = assert_not_none(result.get("fields"), "results/fields")
            if (inj_id := fields.get("InjectionId")) is None:
                msg = (
                    f"Expected InjectionId in 'fields' for result: {fields['ResultId']}"
                )
                raise AssertionError(msg)
            for key, value in fields.items():
                if (
                    key in id_to_injection[inj_id]
                    and id_to_injection[inj_id][key] != value
                ):
                    msg = f"Mismatch between injection field and result field for key: {key}"
                    raise AssertionError(msg)
                id_to_injection[inj_id][key] = value
            if "peaks" in result:
                id_to_injection[inj_id]["peaks"] = [
                    peak["fields"] for peak in result["peaks"]
                ]
            if "curves" in result:
                id_to_injection[inj_id]["curves"] = result["curves"]

        self.injections = list(id_to_injection.values())



from allotropy.allotrope.schema_mappers.adm.liquid_chromatography.benchling._2023._09.liquid_chromatography import (
    MeasurementGroup,
)
from allotropy.parsers.cytiva_unicorn.reader.unicorn_zip_handler import (
    UnicornZipHandler,
)
from allotropy.parsers.cytiva_unicorn.structure.measurements.absorbance import (
    AbsorbanceMeasurement1,
    AbsorbanceMeasurement2,
    AbsorbanceMeasurement3,
)
from allotropy.parsers.cytiva_unicorn.structure.measurements.concentration import (
    ConcentrationMeasurement,
)
from allotropy.parsers.cytiva_unicorn.structure.measurements.conductivity import (
    ConductivityMeasurement,
)
from allotropy.parsers.cytiva_unicorn.structure.measurements.flow import (
    FlowMeasurement,
)
from allotropy.parsers.cytiva_unicorn.structure.measurements.ph import (
    PhMeasurement,
)
from allotropy.parsers.cytiva_unicorn.structure.measurements.pressure import (
    PressureMeasurement,
)
from allotropy.parsers.cytiva_unicorn.structure.measurements.temperature import (
    TemperatureMeasurement,
)
from allotropy.parsers.cytiva_unicorn.structure.static_docs import (
    StaticDocs,
)
from allotropy.parsers.utils.strict_xml_element import (
    StrictXmlElement,
)


def create_measurement_groups(
    handler: UnicornZipHandler, results: StrictXmlElement
) -> list[MeasurementGroup]:
    chrom_1 = handler.get_chrom_1()
    curves = chrom_1.find("Curves")
    elements = curves.findall("Curve")

    static_docs = StaticDocs.create(handler, curves.find("Curve"), results)
    measurements = [
        AbsorbanceMeasurement1.create_or_none(handler, elements, static_docs),
        AbsorbanceMeasurement2.create_or_none(handler, elements, static_docs),
        AbsorbanceMeasurement3.create_or_none(handler, elements, static_docs),
        ConductivityMeasurement.create_or_none(handler, elements, static_docs),
        PhMeasurement.create_or_none(handler, elements, static_docs),
        ConcentrationMeasurement.create_or_none(handler, elements, static_docs),
        PressureMeasurement.create_or_none(handler, elements, static_docs),
        FlowMeasurement.create_or_none(handler, elements, static_docs),
        TemperatureMeasurement.create_or_none(handler, elements, static_docs),
    ]

    return [
        MeasurementGroup(
            measurements=[measurement for measurement in measurements if measurement],
        )
    ]



from allotropy.allotrope.models.adm.plate_reader.benchling._2023._09.plate_reader import (
    Model,
)
from allotropy.allotrope.schema_mappers.adm.plate_reader.benchling._2023._09.plate_reader import (
    Data,
    Mapper,
)
from allotropy.named_file_contents import NamedFileContents
from allotropy.parsers.ctl_immunospot.ctl_immunospot_reader import CtlImmunospotReader
from allotropy.parsers.ctl_immunospot.ctl_immunospot_structure import (
    create_measurement_groups,
    create_metadata,
)
from allotropy.parsers.release_state import ReleaseState
from allotropy.parsers.vendor_parser import VendorParser


class CtlImmunospotParser(VendorParser[Data, Model]):
    DISPLAY_NAME = "CTL ImmunoSpot"
    RELEASE_STATE = ReleaseState.RECOMMENDED
    SUPPORTED_EXTENSIONS = CtlImmunospotReader.SUPPORTED_EXTENSIONS
    SCHEMA_MAPPER = Mapper

    def create_data(self, named_file_contents: NamedFileContents) -> Data:
        reader = CtlImmunospotReader(named_file_contents)
        return Data(
            create_metadata(reader.header),
            create_measurement_groups(
                reader.header,
                reader.plate_identifier,
                reader.plate_data,
                reader.histograms,
            ),
        )



from allotropy.allotrope.models.shared.definitions.definitions import (
    FieldComponentDatatype,
)
from allotropy.allotrope.schema_mappers.adm.liquid_chromatography.benchling._2023._09.liquid_chromatography import (
    DeviceControlDoc,
)
from allotropy.allotrope.schema_mappers.data_cube import DataCube, DataCubeComponent
from allotropy.parsers.cytiva_unicorn.constants import DEVICE_TYPE
from allotropy.parsers.cytiva_unicorn.reader.unicorn_zip_handler import (
    UnicornZipHandler,
)
from allotropy.parsers.cytiva_unicorn.structure.measurements.generic import (
    UnicornMeasurement,
)
from allotropy.parsers.cytiva_unicorn.structure.static_docs import (
    StaticDocs,
)
from allotropy.parsers.utils.strict_xml_element import (
    StrictXmlElement,
)


class FlowMeasurement(UnicornMeasurement):
    @classmethod
    def create_device_control(
        cls,
        device_type: str,
        start_time: str | None,
        sample_flow_data_cube: DataCube | None = None,
        system_flow_data_cube: DataCube | None = None,
    ) -> DeviceControlDoc | None:
        if sample_flow_data_cube is None and system_flow_data_cube is None:
            return None
        return DeviceControlDoc(
            device_type=device_type,
            start_time=start_time,
            sample_flow_data_cube=sample_flow_data_cube,
            system_flow_data_cube=system_flow_data_cube,
        )

    @classmethod
    def create_or_none(
        cls,
        handler: UnicornZipHandler,
        elements: list[StrictXmlElement],
        static_docs: StaticDocs,
    ) -> UnicornMeasurement | None:
        device_controls = [
            cls.create_device_control(
                device_type=DEVICE_TYPE,
                start_time=static_docs.start_time,
                sample_flow_data_cube=cls.get_data_cube_or_none(
                    handler,
                    cls.filter_curve_or_none(elements, r"^Sample flow \(CV/h\)$"),
                    DataCubeComponent(
                        type_=FieldComponentDatatype.float,
                        concept="sample flow",
                        unit="CV/h",
                    ),
                ),
                system_flow_data_cube=cls.get_data_cube_or_none(
                    handler,
                    cls.filter_curve_or_none(elements, r"^System flow \(CV/h\)$"),
                    DataCubeComponent(
                        type_=FieldComponentDatatype.float,
                        concept="system flow",
                        unit="CV/h",
                    ),
                ),
            ),
            cls.create_device_control(
                device_type=DEVICE_TYPE,
                start_time=static_docs.start_time,
                sample_flow_data_cube=cls.get_data_cube_or_none(
                    handler,
                    cls.filter_curve_or_none(elements, r"^Sample flow$"),
                    DataCubeComponent(
                        type_=FieldComponentDatatype.float,
                        concept="sample flow",
                        unit="mL/min",
                    ),
                ),
                system_flow_data_cube=cls.get_data_cube_or_none(
                    handler,
                    cls.filter_curve_or_none(elements, r"^System flow$"),
                    DataCubeComponent(
                        type_=FieldComponentDatatype.float,
                        concept="system flow",
                        unit="mL/min",
                    ),
                ),
            ),
            cls.create_device_control(
                device_type=DEVICE_TYPE,
                start_time=static_docs.start_time,
                sample_flow_data_cube=cls.get_data_cube_or_none(
                    handler,
                    cls.filter_curve_or_none(elements, r"^Sample linear flow$"),
                    DataCubeComponent(
                        type_=FieldComponentDatatype.float,
                        concept="sample flow",
                        unit="cm/s",
                    ),
                ),
            ),
        ]

        measurement = cls.get_measurement(
            static_docs=static_docs,
            device_control_docs=[
                device_control for device_control in device_controls if device_control
            ],
        )
        return measurement if cls.is_valid(cls.get_data_cubes(measurement)) else None

    @classmethod
    def get_data_cubes(cls, measurement: UnicornMeasurement) -> list[DataCube | None]:
        data_cubes = []
        for device_control in measurement.device_control_docs:
            data_cubes.append(device_control.sample_flow_data_cube)
            data_cubes.append(device_control.system_flow_data_cube)
        return data_cubes



from dataclasses import dataclass
from typing import Any

from allotropy.allotrope.converter import add_custom_information_document
from allotropy.allotrope.models.adm.liquid_chromatography.benchling._2023._09.liquid_chromatography import (
    ChromatogramDataCube,
    ChromatographyColumnDocument,
    DataSystemDocument,
    DerivedColumnPressureDataCube,
    DeviceControlAggregateDocument,
    DeviceControlDocumentItem,
    DeviceSystemDocument,
    InjectionDocument,
    LiquidChromatographyAggregateDocument,
    LiquidChromatographyDocumentItem,
    MeasurementAggregateDocument,
    MeasurementDocument,
    Model,
    Peak as PeakDocument,
    PeakList,
    PostColumnPressureDataCube,
    PreColumnPressureDataCube,
    ProcessedDataAggregateDocument,
    ProcessedDataDocumentItem,
    SampleDocument,
    SampleFlowRateDataCube,
    SamplePressureDataCube,
    SolventConcentrationDataCube,
    SystemFlowRateDataCube,
    SystemPressureDataCube,
    TemperatureProfileDataCube,
)
from allotropy.allotrope.models.shared.definitions.custom import (
    TQuantityValueCubicMillimeter,
    TQuantityValueMicrometer,
    TQuantityValueMilliAbsorbanceUnit,
    TQuantityValueMillimeter,
    TQuantityValuePercent,
    TQuantityValueSecondTime,
    TQuantityValueUnitless,
)
from allotropy.allotrope.models.shared.definitions.definitions import TDatacube
from allotropy.allotrope.schema_mappers.data_cube import (
    DataCube,
    get_data_cube,
)
from allotropy.allotrope.schema_mappers.schema_mapper import SchemaMapper
from allotropy.constants import ASM_CONVERTER_VERSION
from allotropy.exceptions import AllotropeConversionError
from allotropy.parsers.utils.values import quantity_or_none, quantity_or_none_from_unit


@dataclass(frozen=True)
class Metadata:
    asset_management_identifier: str
    analyst: str | None = None
    detection_type: str | None = None
    model_number: str | None = None
    software_name: str | None = None
    file_name: str | None = None
    unc_path: str | None = None
    equipment_serial_number: str | None = None
    software_version: str | None = None
    product_manufacturer: str | None = None
    brand_name: str | None = None
    device_identifier: str | None = None
    firmware_version: str | None = None
    description: str | None = None


@dataclass(frozen=True)
class Peak:
    identifier: str
    index: str | None = None
    start: float | None = None
    start_unit: str | None = None
    end: float | None = None
    end_unit: str | None = None
    area: float | None = None
    area_unit: str | None = None
    relative_area: float | None = None
    width: float | None = None
    width_unit: str | None = None
    relative_width: float | None = None
    height: float | None = None
    relative_height: float | None = None
    retention_time: float | None = None
    written_name: str | None = None
    chromatographic_resolution: float | None = None
    chromatographic_asymmetry: float | None = None
    width_at_half_height: float | None = None
    width_at_half_height_unit: str | None = None
    custom_info: dict[str, Any] | None = None


@dataclass(frozen=True)
class DeviceControlDoc:
    device_type: str
    start_time: str | None = None
    solvent_conc_data_cube: DataCube | None = None
    pre_column_pressure_data_cube: DataCube | None = None
    sample_pressure_data_cube: DataCube | None = None
    system_pressure_data_cube: DataCube | None = None
    post_column_pressure_data_cube: DataCube | None = None
    sample_flow_data_cube: DataCube | None = None
    system_flow_data_cube: DataCube | None = None
    temperature_profile_data_cube: DataCube | None = None


@dataclass(frozen=True)
class Measurement:
    measurement_identifier: str
    sample_identifier: str

    # Injection metadata
    injection_identifier: str
    injection_time: str
    autosampler_injection_volume_setting: float

    device_control_docs: list[DeviceControlDoc]

    # Optional metadata
    sample_role_type: str | None = None
    written_name: str | None = None
    chromatography_serial_num: str | None = None
    column_inner_diameter: float | None = None
    chromatography_chemistry_type: str | None = None
    chromatography_particle_size: float | None = None
    batch_identifier: str | None = None

    # Measurement data cubes
    chromatogram_data_cube: DataCube | None = None
    processed_data_chromatogram_data_cube: DataCube | None = None
    derived_column_pressure_data_cube: DataCube | None = None

    peaks: list[Peak] | None = None

    sample_custom_info: dict[str, Any] | None = None


@dataclass(frozen=True)
class MeasurementGroup:
    measurements: list[Measurement]


@dataclass(frozen=True)
class Data:
    metadata: Metadata
    measurement_groups: list[MeasurementGroup]


class Mapper(SchemaMapper[Data, Model]):
    MANIFEST = "http://purl.allotrope.org/manifests/liquid-chromatography/BENCHLING/2023/09/liquid-chromatography.manifest"

    def map_model(self, data: Data) -> Model:
        return Model(
            liquid_chromatography_aggregate_document=LiquidChromatographyAggregateDocument(
                liquid_chromatography_document=[
                    self._get_technique_document(group, data.metadata)
                    for group in data.measurement_groups
                ],
                device_system_document=DeviceSystemDocument(
                    asset_management_identifier=data.metadata.asset_management_identifier,
                    product_manufacturer=data.metadata.product_manufacturer,
                    device_identifier=data.metadata.device_identifier,
                    firmware_version=data.metadata.firmware_version,
                ),
                data_system_document=DataSystemDocument(
                    file_name=data.metadata.file_name,
                    UNC_path=data.metadata.unc_path,
                    software_name=data.metadata.software_name,
                    software_version=data.metadata.software_version,
                    ASM_converter_name=self.converter_name,
                    ASM_converter_version=ASM_CONVERTER_VERSION,
                ),
            ),
            field_asm_manifest=self.MANIFEST,
        )

    def _get_technique_document(
        self, group: MeasurementGroup, metadata: Metadata
    ) -> LiquidChromatographyDocumentItem:
        return LiquidChromatographyDocumentItem(
            analyst=metadata.analyst,
            measurement_aggregate_document=MeasurementAggregateDocument(
                measurement_document=[
                    self._get_measurement_document_item(measurement)
                    for measurement in group.measurements
                ]
            ),
        )

    def _get_measurement_document_item(
        self, measurement: Measurement
    ) -> MeasurementDocument:
        if len(measurement.device_control_docs) < 1:
            msg = "Expected at least one device control document in measurement."
            raise AllotropeConversionError(msg)

        return MeasurementDocument(
            measurement_identifier=measurement.measurement_identifier,
            chromatography_column_document=self._get_chromatography_column_document(
                measurement
            ),
            sample_document=self._get_sample_document(measurement),
            injection_document=(
                self._get_injection_document(measurement)
                if measurement.injection_identifier
                else None
            ),
            processed_data_aggregate_document=self._get_processed_data_aggregate_document(
                measurement
            ),
            device_control_aggregate_document=DeviceControlAggregateDocument(
                device_control_document=[
                    self._get_device_control_document(device_control_doc)
                    for device_control_doc in measurement.device_control_docs
                ]
            ),
            chromatogram_data_cube=(
                get_data_cube(
                    measurement.chromatogram_data_cube,
                    ChromatogramDataCube,
                )
            ),
        )

    def _get_chromatography_column_document(
        self, measurement: Measurement
    ) -> ChromatographyColumnDocument:
        return ChromatographyColumnDocument(
            chromatography_column_serial_number=measurement.chromatography_serial_num,
            chromatography_column_chemistry_type=measurement.chromatography_chemistry_type,
            column_inner_diameter=quantity_or_none(
                TQuantityValueMillimeter,
                measurement.column_inner_diameter,
            ),
            chromatography_column_particle_size=quantity_or_none(
                TQuantityValueMicrometer, measurement.chromatography_particle_size
            ),
        )

    def _get_injection_document(self, measurement: Measurement) -> InjectionDocument:
        return InjectionDocument(
            injection_identifier=measurement.injection_identifier,
            injection_time=self.get_date_time(measurement.injection_time),
            autosampler_injection_volume_setting__chromatography_=TQuantityValueCubicMillimeter(
                value=measurement.autosampler_injection_volume_setting,
            ),
        )

    def _get_sample_document(self, measurement: Measurement) -> SampleDocument:
        return add_custom_information_document(
            SampleDocument(
                sample_identifier=measurement.sample_identifier,
                batch_identifier=measurement.batch_identifier,
                sample_role_type=measurement.sample_role_type,
                written_name=measurement.written_name,
            ),
            measurement.sample_custom_info,
        )

    def _get_peak_document(self, peak: Peak) -> PeakDocument:
        return add_custom_information_document(
            PeakDocument(
                identifier=peak.identifier,
                peak_index=peak.index,
                peak_start=quantity_or_none_from_unit(peak.start_unit, peak.start),  # type: ignore[arg-type]
                peak_end=quantity_or_none_from_unit(peak.end_unit, peak.end),  # type: ignore[arg-type]
                peak_area=quantity_or_none_from_unit(peak.area_unit, peak.area),
                peak_width=quantity_or_none_from_unit(peak.width_unit, peak.width),  # type: ignore[arg-type]
                peak_height=quantity_or_none(
                    TQuantityValueMilliAbsorbanceUnit, peak.height
                ),
                relative_peak_area=quantity_or_none(
                    TQuantityValuePercent, peak.relative_area
                ),
                relative_peak_height=quantity_or_none(
                    TQuantityValuePercent, peak.relative_height
                ),
                retention_time=quantity_or_none(
                    TQuantityValueSecondTime, peak.retention_time
                ),
                written_name=peak.written_name,
                chromatographic_peak_resolution=quantity_or_none(
                    TQuantityValueUnitless, peak.chromatographic_resolution
                ),
                chromatographic_peak_asymmetry_factor=quantity_or_none(
                    TQuantityValueUnitless, peak.chromatographic_asymmetry
                ),
                peak_width_at_half_height=quantity_or_none_from_unit(  # type: ignore[arg-type]
                    peak.width_at_half_height_unit, peak.width_at_half_height
                ),
            ),
            peak.custom_info,
        )

    def _get_processed_data_aggregate_document(
        self, measurement: Measurement
    ) -> ProcessedDataAggregateDocument | None:
        if not (
            measurement.processed_data_chromatogram_data_cube
            or measurement.derived_column_pressure_data_cube
            or measurement.peaks
        ):
            return None
        return ProcessedDataAggregateDocument(
            processed_data_document=[
                ProcessedDataDocumentItem(
                    chromatogram_data_cube=get_data_cube(
                        measurement.processed_data_chromatogram_data_cube,
                        TDatacube,
                    ),
                    derived_column_pressure_data_cube=get_data_cube(
                        measurement.derived_column_pressure_data_cube,
                        DerivedColumnPressureDataCube,
                    ),
                    peak_list=PeakList(
                        peak=[
                            self._get_peak_document(peak) for peak in measurement.peaks
                        ]
                    )
                    if measurement.peaks
                    else None,
                )
            ]
        )

    def _get_device_control_document(
        self, device_control_doc: DeviceControlDoc
    ) -> DeviceControlDocumentItem:
        return DeviceControlDocumentItem(
            device_type=device_control_doc.device_type,
            start_time_setting=(
                self.get_date_time(device_control_doc.start_time)
                if device_control_doc.start_time is not None
                else None
            ),
            solvent_concentration_data_cube=get_data_cube(
                device_control_doc.solvent_conc_data_cube,
                SolventConcentrationDataCube,
            ),
            pre_column_pressure_data_cube=get_data_cube(
                device_control_doc.pre_column_pressure_data_cube,
                PreColumnPressureDataCube,
            ),
            sample_pressure_data_cube=get_data_cube(
                device_control_doc.sample_pressure_data_cube,
                SamplePressureDataCube,
            ),
            system_pressure_data_cube=get_data_cube(
                device_control_doc.system_pressure_data_cube,
                SystemPressureDataCube,
            ),
            post_column_pressure_data_cube=get_data_cube(
                device_control_doc.post_column_pressure_data_cube,
                PostColumnPressureDataCube,
            ),
            sample_flow_rate_data_cube=get_data_cube(
                device_control_doc.sample_flow_data_cube,
                SampleFlowRateDataCube,
            ),
            system_flow_rate_data_cube=get_data_cube(
                device_control_doc.system_flow_data_cube,
                SystemFlowRateDataCube,
            ),
            temperature_profile_data_cube=get_data_cube(
                device_control_doc.temperature_profile_data_cube,
                TemperatureProfileDataCube,
            ),
        )



from enum import Enum

from allotropy.allotrope.schema_mappers.adm.pcr.rec._2024._09.qpcr import ContainerType


# Experiment type was an enum in the BENCHLING/2023/09 version of the qpcr schema, but
# is now just a string field without enum validation. We are keeping this because quantstudio
# adapters have logic around the enum class, but we don't include it in the mapper because is
# not used for validation.
class ExperimentType(str, Enum):
    genotyping_qpcr_experiment = "genotyping qPCR experiment"
    relative_standard_curve_qpcr_experiment = "relative standard curve qPCR experiment"
    melt_curve_qpcr_experiment = "melt curve qPCR experiment"
    comparative_ct_qpcr_experiment = "comparative CT qPCR experiment"
    standard_curve_qpcr_experiment = "standard curve qPCR experiment"
    qpcr_experiment = "qPCR experiment"
    presence_absence_qpcr_experiment = "presence/absence qPCR experiment"
    primary_analysis_experiment = "primary analysis experiment"


DEVICE_TYPE = "qPCR"
PRODUCT_MANUFACTURER = "ThermoFisher Scientific"
DATA_SYSTEM_INSTANCE_IDENTIFIER = "localhost"
CONTAINER_TYPE = ContainerType.PCR_reaction_block



# generated by datamodel-codegen:
#   filename:  spectrophotometry.schema.json
#   timestamp: 2024-09-29T18:51:05+00:00

from __future__ import annotations

from dataclasses import dataclass
from typing import Any

from allotropy.allotrope.models.shared.definitions.custom import (
    TQuantityValueMilliAbsorbanceUnit,
    TQuantityValueMillimeter,
    TQuantityValueNanometer,
    TQuantityValuePicogramPerMilliliter,
    TQuantityValueSecondTime,
    TQuantityValueSquareCentimetersPerGram,
    TQuantityValueSquareCentimetersPerMole,
)
from allotropy.allotrope.models.shared.definitions.definitions import (
    TBooleanValue,
    TClass,
    TDateTimeStampValue,
    TDateTimeValue,
    TDoubleValue,
    TIntegerValue,
    TStatisticDatumRole,
    TStringValue,
    TUnit,
)


@dataclass(kw_only=True)
class AdmCoreREC202406ManifestSchema:
    vocabulary: list[str]
    json_schemas: list[str]
    field_id: str | None = None
    field_type: str | None = None
    shapes: list[str] | None = None


@dataclass(kw_only=True)
class CustomInformationDocumentItem:
    scalar_double_datum: TDoubleValue | None = None
    unit: TUnit | None = None
    scalar_string_datum: TStringValue | None = None
    scalar_timestamp_datum: TDateTimeValue | None = None
    scalar_boolean_datum: TBooleanValue | None = None
    datum_label: TStringValue | None = None


@dataclass(kw_only=True)
class ElectronicProjectRecord:
    written_name: TStringValue
    description: Any | None = None
    start_time: TDateTimeValue | None = None


@dataclass(kw_only=True)
class ErrorDocumentItem:
    error: TStringValue
    error_feature: TStringValue | None = None


@dataclass(kw_only=True)
class ErrorAggregateDocument:
    error_document: list[ErrorDocumentItem] | None = None


@dataclass(kw_only=True)
class ImageDocumentItem:
    experimental_data_identifier: TStringValue | None = None
    index: TIntegerValue | None = None


@dataclass(kw_only=True)
class ImageAggregateDocument:
    image_document: list[ImageDocumentItem] | None = None


@dataclass(kw_only=True)
class StatisticsDocumentItem:
    statistical_feature: TClass


@dataclass(kw_only=True)
class StatisticsAggregateDocument:
    statistics_document: list[StatisticsDocumentItem] | None = None


@dataclass(kw_only=True)
class OrderedItem:
    field_index: int | None = None


@dataclass(kw_only=True)
class TQuantityValueModel:
    value: float
    unit: TUnit
    has_statistic_datum_role: TStatisticDatumRole | None = None
    field_type: TClass | None = None


@dataclass(kw_only=True)
class DataSourceDocumentItem:
    data_source_identifier: TStringValue
    data_source_feature: TStringValue | None = None
    field_index: int | None = None


@dataclass(kw_only=True)
class DataSourceAggregateDocument:
    data_source_document: list[DataSourceDocumentItem]


@dataclass(kw_only=True)
class AnalysisSequenceDocument:
    written_name: TStringValue
    end_time: TDateTimeValue | None = None
    file_name: TStringValue | None = None
    identifier: TStringValue | None = None
    method_identifier: TStringValue | None = None
    method_name: TStringValue | None = None
    start_time: TDateTimeValue | None = None
    UNC_path: TStringValue | None = None
    version_number: TStringValue | None = None


@dataclass(kw_only=True)
class CalculatedDataDocumentItem:
    calculated_data_name: TStringValue
    calculated_result: TQuantityValueModel
    calculated_data_identifier: TStringValue | None = None
    calculation_description: TStringValue | None = None
    data_source_aggregate_document: DataSourceAggregateDocument | None = None
    electronic_project_record: ElectronicProjectRecord | None = None
    field_index: int | None = None


@dataclass(kw_only=True)
class CalculatedDataAggregateDocument:
    calculated_data_document: list[CalculatedDataDocumentItem]


@dataclass(kw_only=True)
class CustomInformationAggregateDocument:
    custom_information_document: list[CustomInformationDocumentItem]


@dataclass(kw_only=True)
class ElectronicSignatureDocumentItem:
    account_identifier: TStringValue
    personal_name: TStringValue
    signature_role_type: TStringValue
    time: TStringValue
    identifier: TStringValue | None = None
    measurement_identifier: TStringValue | None = None
    method_identifier: TStringValue | None = None
    processed_data_identifier: TStringValue | None = None
    field_index: int | None = None


@dataclass(kw_only=True)
class ElectronicSignatureAggregateDocument:
    electronic_signature_document: list[ElectronicSignatureDocumentItem] | None = None


@dataclass(kw_only=True)
class ProcessedDataDocumentItem(OrderedItem):
    custom_information_aggregate_document: CustomInformationAggregateDocument | None = (
        None
    )
    data_processing_document: dict[str, Any] | None = None
    data_source_aggregate_document: DataSourceAggregateDocument | None = None
    electronic_project_record: ElectronicProjectRecord | None = None
    processed_data_identifier: TStringValue | None = None


@dataclass(kw_only=True)
class ProcessedDataAggregateDocument:
    processed_data_document: list[ProcessedDataDocumentItem]
    custom_information_aggregate_document: CustomInformationAggregateDocument | None = (
        None
    )
    electronic_project_record: ElectronicProjectRecord | None = None


@dataclass(kw_only=True)
class DataSystemDocument:
    ASM_file_identifier: TStringValue
    data_system_instance_identifier: TStringValue
    file_name: TStringValue | None = None
    UNC_path: TStringValue | None = None
    ASM_converter_name: TStringValue | None = None
    ASM_converter_version: TStringValue | None = None
    database_primary_key: TStringValue | None = None
    software_name: TStringValue | None = None
    software_version: TStringValue | None = None
    custom_information_aggregate_document: CustomInformationAggregateDocument | None = (
        None
    )


@dataclass(kw_only=True)
class DeviceDocumentItem:
    device_type: TStringValue
    brand_name: TStringValue | None = None
    device_identifier: TStringValue | None = None
    equipment_serial_number: TStringValue | None = None
    firmware_version: TStringValue | None = None
    model_number: TStringValue | None = None
    product_manufacturer: TStringValue | None = None
    custom_information_aggregate_document: CustomInformationAggregateDocument | None = (
        None
    )
    field_index: int | None = None


@dataclass(kw_only=True)
class DeviceSystemDocument:
    device_identifier: TStringValue
    model_number: TStringValue
    asset_management_identifier: TStringValue | None = None
    brand_name: TStringValue | None = None
    description: Any | None = None
    device_document: list[DeviceDocumentItem] | None = None
    equipment_serial_number: TStringValue | None = None
    firmware_version: TStringValue | None = None
    product_manufacturer: TStringValue | None = None
    custom_information_aggregate_document: CustomInformationAggregateDocument | None = (
        None
    )


@dataclass(kw_only=True)
class DiagnosticTraceDocumentItem:
    description: Any
    custom_information_aggregate_document: CustomInformationAggregateDocument | None = (
        None
    )


@dataclass(kw_only=True)
class DiagnosticTraceAggregateDocument:
    diagnostic_trace_document: list[DiagnosticTraceDocumentItem] | None = None
    custom_information_aggregate_document: CustomInformationAggregateDocument | None = (
        None
    )


@dataclass(kw_only=True)
class DeviceControlDocumentItem:
    device_type: TStringValue
    brand_name: TStringValue | None = None
    detection_type: TStringValue | None = None
    device_identifier: TStringValue | None = None
    equipment_serial_number: TStringValue | None = None
    firmware_version: TStringValue | None = None
    model_number: TStringValue | None = None
    product_manufacturer: TStringValue | None = None
    custom_information_aggregate_document: CustomInformationAggregateDocument | None = (
        None
    )
    field_index: int | None = None
    detector_gain_setting: TStringValue | None = None
    exposure_duration_setting: TQuantityValueSecondTime | None = None
    cell_path_length_setting: TQuantityValueMillimeter | None = None
    mass_attenuation_coefficient_setting: TQuantityValueSquareCentimetersPerGram | None = (
        None
    )
    molar_attenuation_coefficient_setting: TQuantityValueSquareCentimetersPerMole | None = (
        None
    )
    detector_wavelength_setting: TQuantityValueNanometer | None = None
    detector_bandwidth_setting: TQuantityValueNanometer | None = None
    electronic_absorbance_wavelength_setting: TQuantityValueNanometer | None = None
    electronic_absorbance_bandwidth_setting: TQuantityValueNanometer | None = None
    electronic_absorbance_reference_bandwidth_setting: TQuantityValueNanometer | None = (
        None
    )
    electronic_absorbance_reference_wavelength_setting: TQuantityValueNanometer | None = (
        None
    )


@dataclass(kw_only=True)
class DeviceControlAggregateDocument:
    device_control_document: list[DeviceControlDocumentItem]
    custom_information_aggregate_document: CustomInformationAggregateDocument | None = (
        None
    )


@dataclass(kw_only=True)
class SampleDocument:
    sample_identifier: TStringValue
    batch_identifier: TStringValue | None = None
    description: Any | None = None
    sample_role_type: TClass | None = None
    written_name: TStringValue | None = None
    custom_information_aggregate_document: CustomInformationAggregateDocument | None = (
        None
    )
    well_location_identifier: TStringValue | None = None
    vial_location_identifier: TStringValue | None = None
    location_identifier: TStringValue | None = None
    well_plate_identifier: TStringValue | None = None
    mass_concentration: TQuantityValuePicogramPerMilliliter | None = None


@dataclass(kw_only=True)
class MeasurementDocument:
    device_control_aggregate_document: DeviceControlAggregateDocument
    measurement_identifier: TStringValue
    sample_document: SampleDocument
    absorbance: TQuantityValueMilliAbsorbanceUnit
    calculated_data_aggregate_document: CalculatedDataAggregateDocument | None = None
    custom_information_aggregate_document: CustomInformationAggregateDocument | None = (
        None
    )
    detection_type: TStringValue | None = None
    electronic_project_record: ElectronicProjectRecord | None = None
    error_aggregate_document: ErrorAggregateDocument | None = None
    image_aggregate_document: ImageAggregateDocument | None = None
    measurement_time: TDateTimeStampValue | None = None
    processed_data_aggregate_document: ProcessedDataAggregateDocument | None = None
    statistics_aggregate_document: StatisticsAggregateDocument | None = None
    mass_concentration: TQuantityValuePicogramPerMilliliter | None = None


@dataclass(kw_only=True)
class MeasurementAggregateDocument:
    measurement_document: list[MeasurementDocument]
    measurement_time: TDateTimeStampValue
    calculated_data_aggregate_document: CalculatedDataAggregateDocument | None = None
    custom_information_aggregate_document: CustomInformationAggregateDocument | None = (
        None
    )
    diagnostic_trace_aggregate_document: DiagnosticTraceAggregateDocument | None = None
    error_aggregate_document: ErrorAggregateDocument | None = None
    image_aggregate_document: ImageAggregateDocument | None = None
    processed_data_aggregate_document: ProcessedDataAggregateDocument | None = None
    statistics_aggregate_document: StatisticsAggregateDocument | None = None
    experimental_data_identifier: TStringValue | None = None
    experiment_type: TStringValue | None = None
    container_type: TClass | None = None


@dataclass(kw_only=True)
class SpectrophotometryDocumentItem:
    measurement_aggregate_document: MeasurementAggregateDocument
    analyst: TStringValue | None = None
    electronic_project_record: ElectronicProjectRecord | None = None
    submitter: TStringValue | None = None


@dataclass(kw_only=True)
class SpectrophotometryAggregateDocument:
    spectrophotometry_document: list[SpectrophotometryDocumentItem]
    analysis_sequence_document: AnalysisSequenceDocument | None = None
    calculated_data_aggregate_document: CalculatedDataAggregateDocument | None = None
    custom_information_aggregate_document: CustomInformationAggregateDocument | None = (
        None
    )
    data_system_document: DataSystemDocument | None = None
    device_system_document: DeviceSystemDocument | None = None
    electronic_project_record: ElectronicProjectRecord | None = None
    electronic_signature_aggregate_document: ElectronicSignatureAggregateDocument | None = (
        None
    )
    processed_data_aggregate_document: ProcessedDataAggregateDocument | None = None
    statistics_aggregate_document: StatisticsAggregateDocument | None = None


@dataclass(kw_only=True)
class Model:
    field_asm_manifest: AdmCoreREC202406ManifestSchema | str
    spectrophotometry_aggregate_document: SpectrophotometryAggregateDocument | None = (
        None
    )



from collections.abc import Iterator, Sequence
from contextlib import contextmanager
from itertools import zip_longest
from pathlib import Path
import shutil

from allotropy.parsers.utils.uuids import random_uuid_str

PathType = Path | str


def _files_equal(path1: PathType, path2: PathType) -> bool:
    with open(str(path1)) as file1, open(str(path2)) as file2:
        for line1, line2 in zip_longest(file1, file2, fillvalue=""):
            if line1 != line2 and not line1.startswith("#   timestamp:"):
                return False
    return True


def _get_backup_path(path: PathType) -> Path:
    _path = Path(path)
    return Path(_path.parent, f".{_path.stem}.{random_uuid_str()}.bak{_path.suffix}")


def get_original_path(path: PathType) -> Path:
    _path = Path(path)
    filename = _path.stem
    if filename.startswith("."):
        # Remove .<uuid>.bak from filename by removing two extensions (by calling .stem)
        filename = f"{Path(Path(filename[1:]).stem).stem}{_path.suffix}"
    return Path(_path.parent, filename)


def is_file_changed(path: PathType) -> bool:
    if is_backup_file(path):
        other_path = get_original_path(path)
    else:
        other_path = _get_backup_path(path)
    if other_path.exists():
        return not _files_equal(path, other_path)
    return True


def _backup_file(path: PathType) -> Path:
    backup_path = _get_backup_path(path)
    if Path(path).exists():
        shutil.copyfile(path, str(backup_path))
    return backup_path


def restore_backup(path: PathType) -> None:
    backup_path = _get_backup_path(path)
    if backup_path.exists():
        backup_path.replace(path)


def is_backup_file(path: PathType) -> bool:
    return ".bak" in Path(path).suffixes


@contextmanager
def backup_paths(
    paths: Sequence[PathType], *, restore: bool | None = False
) -> Iterator[list[Path]]:
    backup_paths = [_backup_file(path) for path in paths]
    try:
        yield backup_paths
    except (Exception, KeyboardInterrupt):
        for backup in backup_paths:
            backup.unlink(missing_ok=True)
        raise

    if not restore:
        for backup, original in zip(backup_paths, paths, strict=True):
            if is_file_changed(backup):
                shutil.copyfile(backup, str(original))
    for backup in backup_paths:
        backup.unlink(missing_ok=True)


@contextmanager
def backup(path: PathType, *, restore: bool | None = False) -> Iterator[Path]:
    with backup_paths([path], restore=restore) as working_path:
        yield working_path[0]



DEFAULT_SOFTWARE_NAME = "xPONENT"
DEFAULT_CONTAINER_TYPE = "well plate"
DEFAULT_DEVICE_TYPE = "multi analyte profiling analyzer"

LUMINEX_EMPTY_PATTERN = r"^[,\"\s]*$"
CALIBRATION_BLOCK_HEADER = "Most Recent Calibration and Verification Results"
TABLE_HEADER_PATTERN = '^"?DataType:"?,"?([^,"]*)"?'
MINIMUM_CALIBRATION_LINE_COLS = 2
EXPECTED_CALIBRATION_RESULT_LEN = 2
EXPECTED_HEADER_COLUMNS = 7
EXPECTED_SECTIONS = ["Median", "Count", "Units", "Dilution Factor", "Warnings/Errors"]



from io import StringIO

import pandas as pd

from allotropy.named_file_contents import NamedFileContents
from allotropy.parsers.lines_reader import CsvReader, read_to_lines
from allotropy.parsers.utils.pandas import read_csv, SeriesData
from allotropy.parsers.utils.values import assert_not_none


class BmgMarsReader:
    SUPPORTED_EXTENSIONS = "csv"
    header: SeriesData
    data: pd.DataFrame
    header_content: str

    def __init__(self, named_file_contents: NamedFileContents) -> None:
        reader = CsvReader(read_to_lines(named_file_contents))
        lines = list(reader.pop_until_inclusive("^,?Raw Data"))
        # Store the header contents so we can parse some values that don't have key/value
        # pairs such as wavelengths and read type.
        self.header_content = "\n".join(lines)

        # Transform header data into a single series
        raw_data = read_csv(StringIO(self.header_content), header=None)
        df = pd.melt(raw_data, value_vars=raw_data.columns.to_list()).dropna(
            axis="index"
        )
        new = df["value"].str.split(": ", expand=True, n=1)
        self.header = SeriesData(pd.Series(new[1].values, index=new[0].str.upper()))

        # Read in the rest of the file as a dataframe
        reader.drop_empty(r"^[,\"\s]*$")
        self.data = assert_not_none(
            reader.pop_csv_block_as_df(header=0, index_col=0),
            msg="Unable to parse dataset from file.",
        )



from allotropy.allotrope.models.adm.solution_analyzer.rec._2024._09.solution_analyzer import (
    Model,
)
from allotropy.allotrope.schema_mappers.adm.solution_analyzer.rec._2024._09.solution_analyzer import (
    Data,
    Mapper,
)
from allotropy.named_file_contents import NamedFileContents
from allotropy.parsers.release_state import ReleaseState
from allotropy.parsers.roche_cedex_bioht.roche_cedex_bioht_reader import (
    RocheCedexBiohtReader,
)
from allotropy.parsers.roche_cedex_bioht.roche_cedex_bioht_structure import (
    create_measurement_groups,
    create_metadata,
    Sample,
    Title,
)
from allotropy.parsers.vendor_parser import VendorParser


class RocheCedexBiohtParser(VendorParser[Data, Model]):
    DISPLAY_NAME = "Roche Cedex BioHT"
    RELEASE_STATE = ReleaseState.RECOMMENDED
    SUPPORTED_EXTENSIONS = RocheCedexBiohtReader.SUPPORTED_EXTENSIONS
    SCHEMA_MAPPER = Mapper

    def create_data(self, named_file_contents: NamedFileContents) -> Data:
        reader = RocheCedexBiohtReader(named_file_contents.contents)

        title = Title.create(reader.title_data)
        samples = Sample.create_samples(reader.samples_data)

        return Data(
            create_metadata(title, named_file_contents.original_file_path),
            measurement_groups=create_measurement_groups(samples, title),
        )



from enum import Enum

from allotropy.allotrope.schema_mappers.adm.pcr.rec._2024._09.qpcr import (
    ContainerType,
    SampleRoleType,
)


class ExperimentType(str, Enum):
    genotyping_qpcr_experiment = "genotyping qPCR experiment"
    relative_standard_curve_qpcr_experiment = "relative standard curve qPCR experiment"
    melt_curve_qpcr_experiment = "melt curve qPCR experiment"
    comparative_ct_qpcr_experiment = "comparative CT qPCR experiment"
    standard_curve_qpcr_experiment = "standard curve qPCR experiment"
    qpcr_experiment = "qPCR experiment"
    presence_absence_qpcr_experiment = "presence/absence qPCR experiment"
    primary_analysis_experiment = "primary analysis experiment"


SAMPLE_ROLE_TYPES_MAP = {
    "NTC": SampleRoleType.control_sample_role,
    "STANDARD": SampleRoleType.standard_sample_role,
    "UNKNOWN": SampleRoleType.unknown_sample_role,
    "POSITIVE_CONTROL": SampleRoleType.control_sample_role,
    "IPC": SampleRoleType.control_sample_role,
    "BLOCKED_IPC": SampleRoleType.control_sample_role,
    "BLOCKEDIPC": SampleRoleType.control_sample_role,
    "PC_ALLELE_1": SampleRoleType.control_sample_role,
    "PC_ALLELE_2": SampleRoleType.control_sample_role,
    "PC_ALLELE_BOTH": SampleRoleType.control_sample_role,
    "POSITIVE_11": SampleRoleType.control_sample_role,
    "POSITIVE_22": SampleRoleType.control_sample_role,
    "POSITIVE_12": SampleRoleType.control_sample_role,
    "Accuracy Control": SampleRoleType.control_sample_role,
}
DEVICE_TYPE = "qPCR"
SOFTWARE_NAME = "Thermo QuantStudio"
SOFTWARE_VERSION = "1.0"
DATA_SYSTEM_INSTANCE_IDENTIFIER = "localhost"
CONTAINER_TYPE = ContainerType.PCR_reaction_block



from __future__ import annotations

from abc import ABC, abstractmethod
from collections.abc import Iterator

from allotropy.calcdocs.extractor import Element
from allotropy.exceptions import AllotropyParserError


def head_tail_dict(
    data: dict[str, str], key: str
) -> tuple[dict[str, str], dict[str, str]]:
    new_keys = data.copy()
    return {key: new_keys.pop(key)}, new_keys


class ViewData:
    def __init__(
        self,
        view: View,
        name: str,
        data: dict[str, ViewData | list[Element]],
    ):
        self.view = view
        self.name = name
        self.data = data

    def filter_keys(self, keys: dict[str, str]) -> dict[str, str]:
        return self.view.filter_keys(keys)

    def iter_keys(self) -> Iterator[dict[str, str]]:
        for key, item in self.data.items():
            full_key = {self.name: key}
            if isinstance(item, ViewData):
                for sub_keys in item.iter_keys():
                    yield {
                        **full_key,
                        **sub_keys,
                    }
            else:
                yield full_key

    def get_item(self, keys: dict[str, str]) -> ViewData | list[Element]:
        if not keys:
            return self

        head, tail = head_tail_dict(keys, self.name)
        item = self.data[head[self.name]]
        return item.get_item(tail) if isinstance(item, ViewData) else item

    def get_sub_view_data(self, keys: dict[str, str]) -> ViewData:
        item = self.get_item(keys)
        if isinstance(item, ViewData):
            return item
        msg = f"Unable to find sub view data with keys: {keys}."
        raise AllotropyParserError(msg)

    def get_leaf_items(self, keys: dict[str, str]) -> list[Element]:
        item = self.get_item(keys)
        if isinstance(item, ViewData):
            msg = f"Unable to find leaf item of view data with keys: {keys}."
            raise AllotropyParserError(msg)
        return item


class View(ABC):
    def __init__(self, name: str, sub_view: View | None = None):
        self.name = name
        self.sub_view = sub_view

    @abstractmethod
    def sort_elements(self, _: list[Element]) -> dict[str, list[Element]]:
        pass

    def filter_keys(self, keys: dict[str, str]) -> dict[str, str]:
        filtered_keys = self.sub_view.filter_keys(keys) if self.sub_view else {}
        key_value = keys.get(self.name)
        if key_value is not None:
            filtered_keys[self.name] = key_value
        return filtered_keys

    def apply(self, elements: list[Element]) -> ViewData:
        return ViewData(
            view=self,
            name=self.name,
            data={
                id_: self.sub_view.apply(element) if self.sub_view else element
                for id_, element in self.sort_elements(elements).items()
            },
        )



from __future__ import annotations

from pathlib import Path

from allotropy.allotrope.schema_mappers.adm.liquid_chromatography.benchling._2023._09.liquid_chromatography import (
    Metadata,
)
from allotropy.parsers.cytiva_unicorn.reader.unicorn_zip_handler import (
    UnicornZipHandler,
)
from allotropy.parsers.utils.strict_xml_element import (
    StrictXmlElement,
)


def get_audit_trail_entry_user(handler: UnicornZipHandler) -> str | None:
    evaluation_log = handler.get_evaluation_log()
    subelement_names = ["AuditTrail", "AuditTrailEntries"]
    if audit_trail_entries := evaluation_log.recursive_find_or_none(subelement_names):
        for element in audit_trail_entries.findall("AuditTrailEntry"):
            if operator := element.find_or_none("Operator"):
                if (name := operator.get_text_or_none()) != "System":
                    return name
    return None


def create_metadata(
    handler: UnicornZipHandler, results: StrictXmlElement, file_path: str
) -> Metadata:
    system_data = handler.get_system_data()
    instrument_config_data = handler.get_instrument_config_data()

    instrument_config = system_data.recursive_find(
        ["System", "InstrumentConfiguration"]
    )

    system_name = results.find_or_none("SystemName")
    firmware_version = instrument_config_data.find_or_none("FirmwareVersion")

    return Metadata(
        asset_management_identifier=instrument_config.get_attr("Description"),
        product_manufacturer="Cytiva Life Sciences",
        device_identifier=system_name.get_text_or_none() if system_name else None,
        firmware_version=(
            firmware_version.get_text_or_none() if firmware_version else None
        ),
        analyst=get_audit_trail_entry_user(handler),
        file_name=Path(file_path).name,
        unc_path=file_path,
    )



from allotropy.allotrope.models.adm.plate_reader.benchling._2023._09.plate_reader import (
    Model,
)
from allotropy.allotrope.schema_mappers.adm.plate_reader.benchling._2023._09.plate_reader import (
    Data,
    Mapper,
)
from allotropy.named_file_contents import NamedFileContents
from allotropy.parsers.lines_reader import CsvReader, read_to_lines
from allotropy.parsers.release_state import ReleaseState
from allotropy.parsers.revvity_kaleido.kaleido_builder import create_data
from allotropy.parsers.revvity_kaleido.kaleido_structure import (
    create_measurement_groups,
    create_metadata,
)
from allotropy.parsers.vendor_parser import VendorParser


class KaleidoParser(VendorParser[Data, Model]):
    DISPLAY_NAME = "Revvity Kaleido"
    RELEASE_STATE = ReleaseState.RECOMMENDED
    SCHEMA_MAPPER = Mapper
    SUPPORTED_EXTENSIONS = "csv"

    def create_data(self, named_file_contents: NamedFileContents) -> Data:
        data = create_data(CsvReader(read_to_lines(named_file_contents)))
        return Data(
            create_metadata(data, named_file_contents.original_file_path),
            create_measurement_groups(data),
        )



from typing import ClassVar

from allotropy.parsers.appbio_quantstudio_designandanalysis.appbio_quantstudio_designandanalysis_reader import (
    DesignQuantstudioReader,
)
from allotropy.parsers.appbio_quantstudio_designandanalysis.constants import (
    ExperimentType,
)
from allotropy.parsers.appbio_quantstudio_designandanalysis.structure.generic.creator import (
    Creator,
)
from allotropy.parsers.appbio_quantstudio_designandanalysis.structure.generic.structure import (
    Data,
    Header,
)
from allotropy.parsers.appbio_quantstudio_designandanalysis.structure.simple_primary_analysis.structure import (
    SimplePrimaryAnalysisWellList,
)


class SimplePrimaryAnalysisCreator(Creator):
    PLUGIN_REGEX: ClassVar[str] = r"^Primary Analysis v\d+\.\d+\.\d+$"
    EXPECTED_SHEETS: ClassVar[list[str]] = ["Results"]

    @classmethod
    def check_sheets(cls, reader: DesignQuantstudioReader) -> bool:
        return list(reader.data.keys()) == cls.EXPECTED_SHEETS

    @classmethod
    def create(cls, reader: DesignQuantstudioReader) -> Data:
        header = Header.create(reader.header)
        wells = SimplePrimaryAnalysisWellList.create(reader, header)
        return Data(
            header,
            wells,
            experiment_type=ExperimentType.primary_analysis_experiment,
            calculated_documents=[],
            reference_target=None,
            reference_sample=None,
        )



import xml.etree.ElementTree as Et

from allotropy.exceptions import (
    AllotropeConversionError,
    AllotropyParserError,
    get_key_or_error,
)
from allotropy.named_file_contents import NamedFileContents
from allotropy.parsers.biorad_bioplex_manager import constants


class BioradBioplexReader:
    def __init__(self, named_file_contents: NamedFileContents) -> None:
        contents = named_file_contents.contents.read()
        try:
            xml_tree = Et.ElementTree(Et.fromstring(contents))  # noqa: S314
            self.root = xml_tree.getroot()
            self.children = {child.tag: child for child in self.root}
        except Et.ParseError as err:
            # Return all expected tags if XML parsing fails
            msg = "Error parsing xml"
            raise AllotropyParserError(msg) from err

        self._validate()

    def _validate(self) -> None:
        missing_tags = [
            tag for tag in constants.EXPECTED_TAGS if tag not in self.children
        ]
        if missing_tags:
            msg = f"Missing expected tags in xml: {missing_tags}"
            raise AllotropeConversionError(msg)

    def __getitem__(self, child_tag: str) -> Et.Element:
        return get_key_or_error("child tag of root", child_tag, self.children)

    def get_attribute(self, child_tag: str, attribute: str) -> str:
        try:
            return self[child_tag].attrib[attribute]
        except KeyError as e:
            msg = f"Unable to find '{attribute}' in {self[child_tag].attrib}"
            raise AllotropeConversionError(msg) from e



from __future__ import annotations

from allotropy.allotrope.models.adm.plate_reader.rec._2024._06.plate_reader import (
    Model,
)
from allotropy.allotrope.schema_mappers.adm.plate_reader.rec._2024._06.plate_reader import (
    Data,
    Mapper,
)
from allotropy.named_file_contents import NamedFileContents
from allotropy.parsers.methodical_mind.methodical_mind_reader import (
    MethodicalMindReader,
)
from allotropy.parsers.methodical_mind.methodical_mind_structure import (
    create_measurement_groups,
    create_metadata,
    Header,
    PlateData,
)
from allotropy.parsers.release_state import ReleaseState
from allotropy.parsers.vendor_parser import VendorParser


class MethodicalMindParser(VendorParser[Data, Model]):
    DISPLAY_NAME = "MSD Methodical Mind"
    RELEASE_STATE = ReleaseState.RECOMMENDED
    SUPPORTED_EXTENSIONS = "txt"
    SCHEMA_MAPPER = Mapper

    def create_data(self, named_file_contents: NamedFileContents) -> Data:
        reader = MethodicalMindReader(named_file_contents)
        return Data(
            create_metadata(
                Header.create(reader.plate_headers[0]),
                named_file_contents.original_file_path,
            ),
            create_measurement_groups(
                [PlateData.create(header, data) for header, data in reader]
            ),
        )



from allotropy.allotrope.models.adm.solution_analyzer.rec._2024._09.solution_analyzer import (
    Model,
)
from allotropy.allotrope.schema_mappers.adm.solution_analyzer.rec._2024._09.solution_analyzer import (
    Data,
    Mapper,
)
from allotropy.named_file_contents import NamedFileContents
from allotropy.parsers.beckman_pharmspec.beckman_pharmspec_reader import (
    BeckmanPharmspecReader,
)
from allotropy.parsers.beckman_pharmspec.beckman_pharmspec_structure import (
    create_calculated_data,
    create_measurement_groups,
    create_metadata,
    Distribution,
    Header,
)
from allotropy.parsers.release_state import ReleaseState
from allotropy.parsers.vendor_parser import VendorParser


class PharmSpecParser(VendorParser[Data, Model]):
    DISPLAY_NAME = "Beckman Coulter PharmSpec"
    RELEASE_STATE = ReleaseState.RECOMMENDED
    SUPPORTED_EXTENSIONS = BeckmanPharmspecReader.SUPPORTED_EXTENSIONS
    SCHEMA_MAPPER = Mapper

    def create_data(self, named_file_contents: NamedFileContents) -> Data:
        reader = BeckmanPharmspecReader(named_file_contents)
        distributions = Distribution.create_distributions(reader.data)
        header = Header.create(reader.header)

        return Data(
            create_metadata(header, named_file_contents.original_file_path),
            create_measurement_groups(header, distributions),
            create_calculated_data(distributions),
        )



from abc import abstractmethod
from dataclasses import dataclass
from typing import Generic, TypeVar

from allotropy.exceptions import AllotropeConversionError
from allotropy.parsers.utils.values import assert_not_none

T = TypeVar("T")


@dataclass(frozen=True)
class Element:
    data: dict[str, float | str | None]

    def get_or_none(self, key: str) -> float | str | None:
        return self.data.get(key)

    def get_str_or_none(self, key: str) -> str | None:
        value = self.get_or_none(key)
        if value is None or isinstance(value, str):
            return value
        msg = f"unable to parse '{value}' as string"
        raise AllotropeConversionError(msg)

    def get_float_or_none(self, key: str) -> float | None:
        value = self.get_or_none(key)
        if value is None or isinstance(value, float):
            return value
        msg = f"unable to parse '{value}' as float"
        raise AllotropeConversionError(msg)

    def get_str(self, key: str) -> str:
        return assert_not_none(self.get_str_or_none(key), key)

    def get_float(self, key: str) -> float:
        return assert_not_none(self.get_float_or_none(key), key)


class Extractor(Generic[T]):
    @classmethod
    @abstractmethod
    def to_element(cls, obj: T) -> Element:
        pass

    @classmethod
    def get_elements(cls, objects: list[T]) -> list[Element]:
        return [cls.to_element(obj) for obj in objects]



from allotropy.allotrope.models.adm.plate_reader.benchling._2023._09.plate_reader import (
    Model,
)
from allotropy.allotrope.schema_mappers.adm.plate_reader.benchling._2023._09.plate_reader import (
    Data,
    Mapper,
)
from allotropy.exceptions import AllotropeConversionError
from allotropy.named_file_contents import NamedFileContents
from allotropy.parsers.example_weyland_yutani.example_weyland_yutani_reader import (
    ExampleWeylandYutaniReader,
)
from allotropy.parsers.example_weyland_yutani.example_weyland_yutani_structure import (
    BasicAssayInfo,
    create_measurement_group,
    create_metadata,
    Instrument,
    Plate,
)
from allotropy.parsers.release_state import ReleaseState
from allotropy.parsers.vendor_parser import VendorParser


class ExampleWeylandYutaniParser(VendorParser[Data, Model]):
    DISPLAY_NAME = "Example Weyland Yutani"
    RELEASE_STATE = ReleaseState.WORKING_DRAFT
    SUPPORTED_EXTENSIONS = ExampleWeylandYutaniReader.SUPPORTED_EXTENSIONS
    SCHEMA_MAPPER = Mapper

    def create_data(self, named_file_contents: NamedFileContents) -> Data:
        reader = ExampleWeylandYutaniReader(named_file_contents)
        basic_assay_info = BasicAssayInfo.create(reader.bottom)
        instrument = Instrument.create()
        plates = Plate.create(reader.middle)

        if plates[0].number_of_wells is None:
            msg = "Unable to determine the number of wells in the plate."
            raise AllotropeConversionError(msg)

        return Data(
            create_metadata(instrument, named_file_contents.original_file_path),
            [create_measurement_group(plate, basic_assay_info) for plate in plates],
        )



from functools import partial

from allotropy.allotrope.models.adm.plate_reader.rec._2024._06.plate_reader import (
    Model,
)
from allotropy.allotrope.schema_mappers.adm.plate_reader.rec._2024._06.plate_reader import (
    Data,
    Mapper,
)
from allotropy.named_file_contents import NamedFileContents
from allotropy.parsers.bmg_labtech_smart_control.bmg_labtech_smart_control_reader import (
    BmgLabtechSmartControlReader,
)
from allotropy.parsers.bmg_labtech_smart_control.bmg_labtech_smart_control_structure import (
    create_calculated_data_documents,
    create_metadata,
    map_measurement_group,
)
from allotropy.parsers.bmg_labtech_smart_control.constants import DISPLAY_NAME
from allotropy.parsers.release_state import ReleaseState
from allotropy.parsers.utils.pandas import map_rows
from allotropy.parsers.vendor_parser import VendorParser


class BmgLabtechSmartControlParser(VendorParser[Data, Model]):
    DISPLAY_NAME = DISPLAY_NAME
    RELEASE_STATE = ReleaseState.RECOMMENDED
    SUPPORTED_EXTENSIONS = BmgLabtechSmartControlReader.SUPPORTED_EXTENSIONS
    SCHEMA_MAPPER = Mapper

    def create_data(self, named_file_contents: NamedFileContents) -> Data:
        reader = BmgLabtechSmartControlReader(named_file_contents)
        measurement_groups = map_rows(
            reader.data, partial(map_measurement_group, headers=reader.header)
        )
        return Data(
            create_metadata(named_file_contents.original_file_path),
            measurement_groups,
            create_calculated_data_documents(measurement_groups, reader),
        )



from dataclasses import dataclass
from typing import Any

from allotropy.allotrope.models.adm.solution_analyzer.rec._2024._03.solution_analyzer import (
    AnalyteAggregateDocument,
    AnalyteDocument,
    DataProcessingDocument,
    DataSystemDocument,
    DeviceControlAggregateDocument,
    DeviceControlDocumentItem,
    DeviceSystemDocument,
    ErrorAggregateDocument,
    ErrorDocumentItem,
    MeasurementAggregateDocument,
    MeasurementDocument,
    Model,
    ProcessedDataAggregateDocument,
    ProcessedDataDocumentItem,
    SampleDocument,
    SolutionAnalyzerAggregateDocument,
    SolutionAnalyzerDocumentItem,
)
from allotropy.allotrope.models.shared.definitions.custom import (
    TQuantityValueCell,
    TQuantityValueDegreeCelsius,
    TQuantityValueGramPerLiter,
    TQuantityValueMicrometer,
    TQuantityValueMilliAbsorbanceUnit,
    TQuantityValueMilliliterPerLiter,
    TQuantityValueMillimeterOfMercury,
    TQuantityValueMillimolePerLiter,
    TQuantityValueMillionCellsPerMilliliter,
    TQuantityValueMilliOsmolesPerKilogram,
    TQuantityValuePercent,
    TQuantityValuePH,
    TQuantityValueUnitless,
)
from allotropy.allotrope.schema_mappers.schema_mapper import SchemaMapper
from allotropy.constants import ASM_CONVERTER_VERSION
from allotropy.exceptions import AllotropeConversionError
from allotropy.parsers.utils.values import quantity_or_none


@dataclass(frozen=True)
class Analyte:
    name: str
    value: float
    unit: str

    def __lt__(self, other: Any) -> bool:
        if not isinstance(other, Analyte):
            return False

        return self.name < other.name


@dataclass(frozen=True)
class Error:
    error: str
    feature: str | None = None


@dataclass(frozen=True)
class Measurement:
    # Measurement metadata
    identifier: str
    measurement_time: str
    sample_identifier: str
    description: str | None = None
    batch_identifier: str | None = None
    detection_type: str | None = None

    # Measurements
    absorbance: float | None = None
    po2: float | None = None
    pco2: float | None = None
    carbon_dioxide_saturation: float | None = None
    oxygen_saturation: float | None = None
    ph: float | None = None
    temperature: float | None = None
    osmolality: float | None = None
    viability: float | None = None
    total_cell_density: float | None = None
    viable_cell_density: float | None = None
    average_live_cell_diameter: float | None = None
    total_cell_count: float | None = None
    viable_cell_count: float | None = None
    cell_type_processing_method: str | None = None
    cell_density_dilution_factor: float | None = None
    analytes: list[Analyte] | None = None

    # Errors
    errors: list[Error] | None = None


@dataclass(frozen=True)
class MeasurementGroup:
    measurements: list[Measurement]
    analyst: str | None
    data_processing_time: str | None


@dataclass(frozen=True)
class Metadata:
    device_type: str
    device_identifier: str | None = None
    model_number: str | None = None
    software_name: str | None = None
    detection_type: str | None = None
    unc_path: str | None = None
    software_version: str | None = None
    equipment_serial_number: str | None = None
    product_manufacturer: str | None = None
    brand_name: str | None = None

    file_name: str | None = None
    data_system_instance_identifier: str | None = None

    analyst: str | None = None
    measurement_time: str | None = None
    analytical_method_identifier: str | None = None
    method_version: str | None = None
    experimental_data_identifier: str | None = None


@dataclass(frozen=True)
class Data:
    metadata: Metadata
    measurement_groups: list[MeasurementGroup]


class Mapper(SchemaMapper[Data, Model]):
    MANIFEST = "http://purl.allotrope.org/manifests/solution-analyzer/REC/2024/03/solution-analyzer.manifest"

    def map_model(self, data: Data) -> Model:
        return Model(
            solution_analyzer_aggregate_document=SolutionAnalyzerAggregateDocument(
                device_system_document=DeviceSystemDocument(
                    model_number=data.metadata.model_number,
                    equipment_serial_number=data.metadata.equipment_serial_number,
                    device_identifier=data.metadata.device_identifier,
                    product_manufacturer=data.metadata.product_manufacturer,
                ),
                data_system_document=DataSystemDocument(
                    file_name=data.metadata.file_name,
                    UNC_path=data.metadata.unc_path,
                    software_name=data.metadata.software_name,
                    software_version=data.metadata.software_version,
                    ASM_converter_name=self.converter_name,
                    ASM_converter_version=ASM_CONVERTER_VERSION,
                ),
                solution_analyzer_document=[
                    self._get_technique_document(measurement_group, data.metadata)
                    for measurement_group in data.measurement_groups
                ],
            ),
            field_asm_manifest=self.MANIFEST,
        )

    def _get_technique_document(
        self, measurement_group: MeasurementGroup, metadata: Metadata
    ) -> SolutionAnalyzerDocumentItem:
        return SolutionAnalyzerDocumentItem(
            analyst=measurement_group.analyst,
            measurement_aggregate_document=MeasurementAggregateDocument(
                data_processing_time=self.get_date_time(
                    measurement_group.data_processing_time
                )
                if measurement_group.data_processing_time
                else None,
                measurement_document=[
                    self._get_measurement_document_item(measurement, metadata)
                    for measurement in measurement_group.measurements
                ],
            ),
        )

    def _get_measurement_document_item(
        self, measurement: Measurement, metadata: Metadata
    ) -> MeasurementDocument:
        return MeasurementDocument(
            measurement_identifier=measurement.identifier,
            measurement_time=self.get_date_time(measurement.measurement_time),
            sample_document=self._get_sample_document(measurement),
            device_control_aggregate_document=DeviceControlAggregateDocument(
                device_control_document=[
                    DeviceControlDocumentItem(
                        device_type=metadata.device_type,
                        detection_type=measurement.detection_type,
                    ),
                ]
            ),
            analyte_aggregate_document=AnalyteAggregateDocument(
                analyte_document=[
                    self._create_analyte_document(analyte)
                    for analyte in measurement.analytes
                ]
            )
            if measurement.analytes
            else None,
            processed_data_aggregate_document=self._create_processed_data_document(
                measurement
            ),
            error_aggregate_document=self._get_error_aggregate_document(
                measurement.errors
            ),
            absorbance=quantity_or_none(
                TQuantityValueMilliAbsorbanceUnit, measurement.absorbance
            ),
            pO2=quantity_or_none(TQuantityValueMillimeterOfMercury, measurement.po2),
            pCO2=quantity_or_none(TQuantityValueMillimeterOfMercury, measurement.pco2),
            carbon_dioxide_saturation=quantity_or_none(
                TQuantityValuePercent, measurement.carbon_dioxide_saturation
            ),
            oxygen_saturation=quantity_or_none(
                TQuantityValuePercent, measurement.oxygen_saturation
            ),
            pH=quantity_or_none(TQuantityValuePH, measurement.ph),
            temperature=quantity_or_none(
                TQuantityValueDegreeCelsius, measurement.temperature
            ),
            osmolality=quantity_or_none(
                TQuantityValueMilliOsmolesPerKilogram, measurement.osmolality
            ),
        )

    def _get_sample_document(self, measurement: Measurement) -> SampleDocument:
        return SampleDocument(
            sample_identifier=measurement.sample_identifier,
            batch_identifier=measurement.batch_identifier,
            description=measurement.description,
        )

    def _create_analyte_document(self, analyte: Analyte) -> AnalyteDocument:
        if analyte.unit == "g/L":
            return AnalyteDocument(
                analyte_name=analyte.name,
                mass_concentration=TQuantityValueGramPerLiter(value=analyte.value),
            )
        elif analyte.unit == "mL/L":
            return AnalyteDocument(
                analyte_name=analyte.name,
                volume_concentration=TQuantityValueMilliliterPerLiter(
                    value=analyte.value
                ),
            )
        elif analyte.unit == "mmol/L":
            return AnalyteDocument(
                analyte_name=analyte.name,
                molar_concentration=TQuantityValueMillimolePerLiter(
                    value=analyte.value
                ),
            )
        elif analyte.unit == "U/L":
            if analyte.name == "ldh":
                return AnalyteDocument(
                    analyte_name=analyte.name,
                    molar_concentration=TQuantityValueMillimolePerLiter(
                        value=analyte.value * 0.0167
                        if analyte.value > 0
                        else analyte.value
                    ),
                )
            else:
                msg = f"Invalid unit for {analyte.name}: {analyte.unit}"
                raise AllotropeConversionError(msg)

        msg = f"Invalid unit for analyte: {analyte.unit}, value values are: g/L, mL/L, mmol/L"
        raise AllotropeConversionError(msg)

    def _create_processed_data_document(
        self, measurement: Measurement
    ) -> ProcessedDataAggregateDocument | None:
        processed_data_document = ProcessedDataDocumentItem(
            viability__cell_counter_=quantity_or_none(
                TQuantityValuePercent, measurement.viability
            ),
            total_cell_density__cell_counter_=quantity_or_none(
                TQuantityValueMillionCellsPerMilliliter, measurement.total_cell_density
            ),
            viable_cell_density__cell_counter_=quantity_or_none(
                TQuantityValueMillionCellsPerMilliliter, measurement.viable_cell_density
            ),
            average_live_cell_diameter__cell_counter_=quantity_or_none(
                TQuantityValueMicrometer, measurement.average_live_cell_diameter
            ),
            total_cell_count=quantity_or_none(
                TQuantityValueCell, measurement.total_cell_count
            ),
            viable_cell_count=quantity_or_none(
                TQuantityValueCell, measurement.viable_cell_count
            ),
            data_processing_document=DataProcessingDocument(
                cell_type_processing_method=measurement.cell_type_processing_method,
                cell_density_dilution_factor=quantity_or_none(
                    TQuantityValueUnitless, measurement.cell_density_dilution_factor
                ),
            )
            if measurement.cell_type_processing_method
            or measurement.cell_density_dilution_factor is not None
            else None,
        )

        if all(value is None for value in processed_data_document.__dict__.values()):
            return None

        return ProcessedDataAggregateDocument(
            processed_data_document=[processed_data_document]
        )

    def _get_error_aggregate_document(
        self, errors: list[Error] | None
    ) -> ErrorAggregateDocument | None:
        if not errors:
            return None

        return ErrorAggregateDocument(
            error_document=[
                ErrorDocumentItem(error=error.error, error_feature=error.feature)
                for error in errors
            ]
        )



""" Constants file for ThermoFisher Genesys 30 Adapter"""

# Instrument software details
GENESYS_SOFTWARE = "Genesys 30 instrument software"
DISPLAY_NAME = "Thermo Fisher Scientific Genesys30"
MODEL_NUMBER = "Genesys 30"
BRAND_NAME = "Genesys"
PRODUCT_MANUFACTURER = "Thermo Fisher Scientific"
DEVICE_TYPE = "absorbance detector"

# Error messages:
MEASUREMENT_TIME_MISSING = "Unable to find measurement time from instrument file"



from abc import abstractmethod
from collections.abc import Iterator
import struct


class Converter:
    @abstractmethod
    def get_format(self) -> str:
        pass

    @abstractmethod
    def iter_data(self, data: bytes) -> Iterator[bytes]:
        pass

    def convert(self, data: bytes) -> tuple[float, ...]:
        binary_format = self.get_format()
        return tuple(
            struct.unpack(binary_format, binary_data)[0]
            for binary_data in self.iter_data(data)
        )


class FloatConverter(Converter):
    def get_format(self) -> str:
        return "<f"  # little endian float (4 bytes)

    def iter_data(self, data: bytes) -> Iterator[bytes]:
        for i in range(47, len(data) - 48, 4):
            yield data[i : i + 4]



from copy import deepcopy
from pathlib import Path

import pandas as pd

from allotropy.allotrope.schema_mappers.adm.liquid_handler.benchling._2024._11.liquid_handler import (
    Device,
    Measurement,
    MeasurementGroup,
    Metadata,
)
from allotropy.parsers.beckman_coulter_biomek import constants
from allotropy.parsers.constants import NOT_APPLICABLE
from allotropy.parsers.utils.pandas import map_rows, SeriesData
from allotropy.parsers.utils.uuids import random_uuid_str


def create_metadata(data: SeriesData, file_path: str) -> Metadata:
    pod_head_serial_columns = [
        str(column) for column in data.series.index if "head serial number" in column
    ]
    devices = [
        Device(
            identifier=pod_head_serial_column.split(" ")[0],
            device_type=constants.PROBE_HEAD_DEVICE_TYPE,
            serial_number=data[str, pod_head_serial_column],
            product_manufacturer=constants.PRODUCT_MANUFACTURER,
        )
        for pod_head_serial_column in pod_head_serial_columns
        if data[str, pod_head_serial_column] != "None"
    ]

    path = Path(file_path)
    return Metadata(
        file_name=path.name,
        asm_file_identifier=path.with_suffix(".json").name,
        unc_path=str(path),
        data_system_instance_identifier=NOT_APPLICABLE,
        device_type=constants.DEVICE_TYPE,
        equipment_serial_number=data.get(str, "Unit serial number"),
        product_manufacturer=constants.PRODUCT_MANUFACTURER,
        software_name=constants.SOFTWARE_NAME,
        devices=devices,
    )


def _create_measurement(
    aspiration_data: SeriesData, dispense_data: SeriesData
) -> Measurement:
    return Measurement(
        identifier=random_uuid_str(),
        measurement_time=dispense_data[str, "Time Stamp"],
        sample_identifier=NOT_APPLICABLE,
        source_plate=aspiration_data[str, "Labware Barcode"],
        source_well=aspiration_data[str, "Well Index"],
        source_location=aspiration_data[str, "Deck Position"],
        destination_plate=dispense_data[str, "Labware Barcode"],
        destination_well=dispense_data[str, "Well Index"],
        destination_location=dispense_data[str, "Deck Position"],
        aspiration_volume=aspiration_data[float, "Amount"],
        transfer_volume=dispense_data[float, "Amount"],
        device_control_custom_info={
            "probe": dispense_data[str, "Probe"],
            "pod": aspiration_data.get(str, "Pod"),
            "source labware name": aspiration_data.get(str, "Labware Name"),
            "destination labware name": dispense_data.get(str, "Labware Name"),
            "source liquid handling technique": aspiration_data.get(
                str, "Liquid Handling Technique"
            ),
            "destination liquid handling technique": dispense_data.get(
                str, "Liquid Handling Technique"
            ),
        },
    )


def create_measurement_groups(
    data: pd.DataFrame, header: SeriesData
) -> list[MeasurementGroup]:
    # This function will be called for every row in the dataset, use it to create
    # a corresponding measurement group.
    measurements: list[Measurement] = []
    probe_to_aspiration: dict[str, SeriesData] = {}

    def map_to_measurements(row_data: SeriesData) -> None:
        probe = row_data[str, "Probe"]
        transfer_step = row_data[str, "Transfer Step"]
        if transfer_step == constants.TransferStep.ASPIRATE.value:
            if probe in probe_to_aspiration:
                msg = f"Got a second Aspirate step before a Transfer step for probe {probe}"
                raise AssertionError(msg)
            probe_to_aspiration[probe] = deepcopy(row_data)
        elif transfer_step == constants.TransferStep.DISPENSE.value:
            if probe not in probe_to_aspiration:
                msg = f"Got a Transfer step before an Aspirate step for probe {probe}"
                raise AssertionError(msg)
            aspiration_data = probe_to_aspiration.pop(probe)
            measurements.append(_create_measurement(aspiration_data, row_data))
        else:
            msg = f"Got unexpected Transfer Step: {transfer_step}"
            raise AssertionError(msg)

    map_rows(data, map_to_measurements)

    return [
        MeasurementGroup(
            analyst=header[str, "Logged in user"],
            analytical_method_identifier=header.get(str, "Method"),
            measurements=measurements,
        )
    ]



from allotropy.allotrope.models.shared.components.plate_reader import SampleRoleType

BRAND_NAME = "Biacore"
PRODUCT_MANUFACTURER = "Cytiva"
DISPLAY_NAME = "Cytiva Biacore T200 Control"
SURFACE_PLASMON_RESONANCE = "surface plasmon resonance"
DEVICE_TYPE = "binding affinity analyzer"

SAMPLE_ROLE_TYPE = {
    "blank role": SampleRoleType.blank_role.value,
    "sample role": SampleRoleType.sample_role.value,
}



import os
from typing import Any

from allotropy.allotrope.schema_mappers.adm.binding_affinity_analyzer.benchling._2024._12.binding_affinity_analyzer import (
    DeviceDocument,
    MeasurementGroup,
    Measurements,
    MeasurementType,
    Metadata,
)
from allotropy.named_file_contents import NamedFileContents
from allotropy.parsers.constants import NOT_APPLICABLE
from allotropy.parsers.cytiva_biacore_t200_control import constants
from allotropy.parsers.utils.uuids import random_uuid_str
from allotropy.parsers.utils.values import (
    assert_not_none,
    try_float_or_none,
    try_int_or_none,
)


def create_metadata(
    intermediate_structured_data: dict[str, Any], named_file_contents: NamedFileContents
) -> Metadata:
    if (
        "sample_data"
        in intermediate_structured_data["application_template_details"].keys()
    ):
        compartment_temperature = intermediate_structured_data[
            "application_template_details"
        ]["RackTemperature"].get("Value")
    else:
        compartment_temperature = intermediate_structured_data[
            "application_template_details"
        ]["system_preparations"].get("RackTemp")
    return Metadata(
        brand_name=constants.BRAND_NAME,
        device_identifier=assert_not_none(
            intermediate_structured_data["system_information"].get("InstrumentId"),
            "InstrumentId",
        ),
        product_manufacturer=constants.PRODUCT_MANUFACTURER,
        asm_file_identifier=f"{os.path.splitext(os.path.basename(named_file_contents.original_file_path))[0]}.json"
        if os.path.basename(named_file_contents.original_file_path)
        else "N/A",
        model_number=assert_not_none(
            intermediate_structured_data["system_information"].get("ProcessingUnit"),
            "ProcessingUnit",
        ),
        data_system_instance_identifier=NOT_APPLICABLE,
        file_name=os.path.basename(named_file_contents.original_file_path),
        unc_path=named_file_contents.original_file_path,
        software_name=intermediate_structured_data["system_information"].get(
            "Application"
        ),
        software_version=intermediate_structured_data["system_information"].get(
            "Version"
        ),
        detection_type=constants.SURFACE_PLASMON_RESONANCE,
        compartment_temperature=try_float_or_none(compartment_temperature)
        if compartment_temperature is not None
        else None,
        sensor_chip_type=intermediate_structured_data["chip"].get("Name"),
        lot_number=None
        if intermediate_structured_data["chip"].get("LotNo") == ""
        else intermediate_structured_data["chip"].get("LotNo"),
        sensor_chip_identifier=assert_not_none(
            intermediate_structured_data["chip"].get("Id"), "Chip ID"
        ),
        device_document=[
            DeviceDocument(
                device_type=key.split()[0], device_identifier=key.split()[-1]
            )
            for key in intermediate_structured_data["application_template_details"]
            if key.startswith("Flowcell")
        ]
        if any(
            key.startswith("Flowcell")
            for key in intermediate_structured_data["application_template_details"]
        )
        else None,
        sensor_chip_custom_info={
            "ifc identifier": intermediate_structured_data["chip"].get("IFC"),
            "last modified time": intermediate_structured_data["chip"].get(
                "LastModTime"
            ),
            "last use time": intermediate_structured_data["chip"].get("LastUseTime"),
            "first dock date": intermediate_structured_data["chip"].get(
                "FirstDockDate"
            ),
        },
    )


def create_measurements(
    intermediate_structured_data: dict[str, Any]
) -> list[Measurements]:
    if "sample_data" in intermediate_structured_data.keys():
        detection_setting = intermediate_structured_data[
            "application_template_details"
        ]["detection"]
        detection_type = detection_setting["Detection"]
        detection_key = f"Detection{detection_type}"
        detection_value = detection_setting[detection_key]
        return [
            Measurements(
                identifier=random_uuid_str(),
                sensorgram_posix_path=assert_not_none(
                    intermediate_structured_data["cycle_data"][i].get(
                        "sensorgram_path"
                    ),
                    "sensorgram path",
                ),
                sensorgram_posix_path_identifier=(
                    os.path.basename(
                        intermediate_structured_data["cycle_data"][i].get(
                            "sensorgram_path"
                        )
                    )
                    if intermediate_structured_data["cycle_data"][i].get(
                        "sensorgram_path"
                    )
                    is not None
                    else None
                ),
                rpoint_posix_path=intermediate_structured_data["cycle_data"][i].get(
                    "r-point_path"
                ),
                rpoint_posix_path_identifier=(
                    os.path.basename(
                        intermediate_structured_data["cycle_data"][i].get(
                            "r-point_path"
                        )
                    )
                    if intermediate_structured_data["cycle_data"][i].get("r-point_path")
                    is not None
                    else None
                ),
                type_=MeasurementType.SURFACE_PLASMON_RESONANCE,
                device_type=constants.DEVICE_TYPE,
                sample_identifier=intermediate_structured_data["sample_data"][i].get(
                    "sample_name", NOT_APPLICABLE
                ),
                location_identifier=intermediate_structured_data["sample_data"][i].get(
                    "rack"
                ),
                sample_role_type=constants.SAMPLE_ROLE_TYPE.get(
                    intermediate_structured_data["sample_data"][i].get("role")
                ),
                concentration=try_float_or_none(
                    intermediate_structured_data["sample_data"][i].get("concentration")
                )
                if "concentration"
                in intermediate_structured_data["sample_data"][i].keys()
                else None,
                device_control_custom_info={
                    "number of flow cells": try_int_or_none(
                        intermediate_structured_data["chip"].get("NoFcs")
                    ),
                    "number of spots": try_int_or_none(
                        intermediate_structured_data["chip"].get("NoSpots")
                    ),
                    "buffer volume": {
                        "value": try_float_or_none(
                            next(
                                value
                                for key, value in intermediate_structured_data[
                                    "application_template_details"
                                ]["prepare_run"].items()
                                if key.startswith("Buffer")
                            )
                        ),
                        "unit": "mL",
                    },
                    **{detection_key.lower(): detection_value},
                },
                sample_custom_info={
                    "molecular weight": {
                        "value": try_float_or_none(
                            intermediate_structured_data["sample_data"][i].get(
                                "molecular_weight"
                            )
                        ),
                        "unit": "Da",
                    }
                },
            )
            for i in range(intermediate_structured_data["total_cycles"])
        ]
    else:
        return [
            Measurements(
                identifier=random_uuid_str(),
                sensorgram_posix_path=assert_not_none(
                    intermediate_structured_data["cycle_data"][i].get(
                        "sensorgram_path"
                    ),
                    "sensorgram path",
                ),
                sensorgram_posix_path_identifier=(
                    os.path.basename(
                        intermediate_structured_data["cycle_data"][i].get(
                            "sensorgram_path"
                        )
                    )
                    if intermediate_structured_data["cycle_data"][i].get(
                        "sensorgram_path"
                    )
                    is not None
                    else None
                ),
                rpoint_posix_path=intermediate_structured_data["cycle_data"][i].get(
                    "r-point_path"
                ),
                rpoint_posix_path_identifier=(
                    os.path.basename(
                        intermediate_structured_data["cycle_data"][i].get(
                            "r-point_path"
                        )
                    )
                    if intermediate_structured_data["cycle_data"][i].get("r-point_path")
                    is not None
                    else None
                ),
                type_=MeasurementType.SURFACE_PLASMON_RESONANCE,
                device_type="binding affinity analyzer",
                sample_identifier=intermediate_structured_data.get(
                    "sample_data", NOT_APPLICABLE
                ),
                method_name=intermediate_structured_data[
                    "application_template_details"
                ][f"Flowcell {i + 1}"].get("MethodName"),
                ligand_identifier=intermediate_structured_data[
                    "application_template_details"
                ][f"Flowcell {i + 1}"].get("Ligand"),
                flow_cell_identifier=f"Flowcell {i + 1}",
                flow_path=intermediate_structured_data["application_template_details"][
                    f"Flowcell {i + 1}"
                ].get("DetectionText"),
                flow_rate=intermediate_structured_data["application_template_details"][
                    f"Flowcell {i + 1}"
                ].get("Flow"),
                contact_time=intermediate_structured_data[
                    "application_template_details"
                ][f"Flowcell {i + 1}"].get("ContactTime"),
                dilution=try_int_or_none(
                    intermediate_structured_data["application_template_details"][
                        f"Flowcell {i + 1}"
                    ].get("DilutePercent")
                ),
                device_control_custom_info={
                    "number of flow cells": try_int_or_none(
                        intermediate_structured_data["chip"].get("NoFcs")
                    ),
                    "number of spots": try_int_or_none(
                        intermediate_structured_data["chip"].get("NoSpots")
                    ),
                    "buffer volume": {
                        "value": try_float_or_none(
                            next(
                                value
                                for key, value in intermediate_structured_data[
                                    "application_template_details"
                                ]["prepare_run"].items()
                                if key.startswith("Buffer")
                            )
                        ),
                        "unit": "mL",
                    },
                },
            )
            for i in range(intermediate_structured_data["total_cycles"])
        ]


def create_measurement_groups(
    intermediate_structured_data: dict[str, Any]
) -> MeasurementGroup:
    return MeasurementGroup(
        measurement_time=assert_not_none(
            intermediate_structured_data["system_information"].get("Timestamp"),
            "Timestamp",
        ),
        measurements=create_measurements(intermediate_structured_data),
        experiment_type=intermediate_structured_data["system_information"].get(
            "RunTypeId"
        ),
        analytical_method_identifier=intermediate_structured_data[
            "system_information"
        ].get("TemplateFile"),
        analyst=intermediate_structured_data["application_template_details"][
            "properties"
        ].get("User"),
        measurement_aggregate_custom_info={
            "baseline flow": {
                "value": try_float_or_none(
                    intermediate_structured_data["application_template_details"][
                        "BaselineFlow"
                    ].get("value")
                )
                if "BaselineFlow"
                in intermediate_structured_data["application_template_details"].keys()
                else None,
                "unit": "μL/min",
            },
            "data collection rate": {
                "value": try_float_or_none(
                    intermediate_structured_data["application_template_details"][
                        "DataCollectionRate"
                    ].get("value")
                )
                if "DataCollectionRate"
                in intermediate_structured_data["application_template_details"].keys()
                else None,
                "unit": "Hz",
            },
            "dip details": [
                {
                    "datum label": "norm_dip_details",
                    "datum value": norm_datum.get("response"),
                    "sweep row": norm_datum.get("sweep_row"),
                    "flow cell identifier": norm_datum.get("flow_cell"),
                }
                for norm_datum in intermediate_structured_data["dip"]["norm_data"]
            ]
            + [
                {
                    "datum label": "raw_dip_details",
                    "datum value": raw_datum.get("response"),
                    "sweep_row": raw_datum.get("sweep_row"),
                    "flow cell identifier": raw_datum.get("flow_cell"),
                }
                for raw_datum in intermediate_structured_data["dip"]["raw_data"]
            ],
        },
    )



DEVICE_IDENTIFIER = "Nanodrop"
DEVICE_TYPE = "absorbance detector"
MODEL_NUBMER = "Nanodrop Eight"

# These may be reported in the results, and are stored as calculated data
ABSORBANCE_RATIOS = [
    (260, 280),
    (260, 230),
]

CONCENTRATION_UNITS = ["ng/μL", "μg/mL", "μg/µL", "mg/mL", "μg,mL"]



from __future__ import annotations

from allotropy.allotrope.models.adm.plate_reader.rec._2024._06.plate_reader import (
    Model,
)
from allotropy.allotrope.schema_mappers.adm.plate_reader.rec._2024._06.plate_reader import (
    Data,
    Mapper,
)
from allotropy.named_file_contents import NamedFileContents
from allotropy.parsers.methodical_mind.methodical_mind_reader import (
    MethodicalMindReader,
)
from allotropy.parsers.methodical_mind.methodical_mind_structure import (
    create_measurement_groups,
    create_metadata,
    Header,
    PlateData,
)
from allotropy.parsers.msd_workbench.msd_workbench_calculated_data_mapping import (
    create_calculated_data_groups,
)
from allotropy.parsers.msd_workbench.msd_workbench_reader import (
    MSDWorkbenchReader,
)
from allotropy.parsers.msd_workbench.msd_workbench_structure import (
    create_measurement_groups as create_msd_measurement_groups,
    create_metadata as create_msd_metadata,
    PlateData as MSDPlateData,
)
from allotropy.parsers.release_state import ReleaseState
from allotropy.parsers.vendor_parser import VendorParser


class MSDWorkbenchParser(VendorParser[Data, Model]):
    DISPLAY_NAME = "MSD Discovery Workbench"
    RELEASE_STATE = ReleaseState.RECOMMENDED
    SUPPORTED_EXTENSIONS = MSDWorkbenchReader.SUPPORTED_EXTENSIONS
    SCHEMA_MAPPER = Mapper

    def create_data(self, named_file_contents: NamedFileContents) -> Data:
        if named_file_contents.extension == "txt":
            return self._process_methodical_mind(named_file_contents)
        return self._process_msd_workbench(named_file_contents)

    def _process_methodical_mind(self, named_file_contents: NamedFileContents) -> Data:
        reader = MethodicalMindReader(named_file_contents)
        metadata = create_metadata(
            Header.create(reader.plate_headers[0]),
            named_file_contents.original_file_path,
        )
        plate_data_objects = [PlateData.create(header, data) for header, data in reader]
        measurement_groups = create_measurement_groups(plate_data_objects)
        return Data(metadata, measurement_groups)

    def _process_msd_workbench(self, named_file_contents: NamedFileContents) -> Data:
        reader = MSDWorkbenchReader(named_file_contents)
        plate_data = MSDPlateData.create(reader.plate_data, reader.well_plate_id)
        measurement_groups = create_msd_measurement_groups(plate_data)
        measurements = [
            measurement
            for group in measurement_groups
            for measurement in group.measurements
        ]
        calculated_data_groups = create_calculated_data_groups(
            reader.plate_data, measurements
        )
        metadata = create_msd_metadata(named_file_contents.original_file_path)
        return Data(metadata, measurement_groups, calculated_data_groups)



from allotropy.allotrope.models.adm.plate_reader.benchling._2023._09.plate_reader import (
    Model,
)
from allotropy.allotrope.schema_mappers.adm.plate_reader.benchling._2023._09.plate_reader import (
    Data,
    Mapper,
)
from allotropy.named_file_contents import NamedFileContents
from allotropy.parsers.agilent_gen5.agilent_gen5_reader import AgilentGen5Reader
from allotropy.parsers.agilent_gen5.agilent_gen5_structure import (
    get_identifiers,
    get_results_section,
    HeaderData,
)
from allotropy.parsers.agilent_gen5_image.agilent_gen5_image_structure import (
    create_metadata,
    create_results,
    ReadData,
)
from allotropy.parsers.release_state import ReleaseState
from allotropy.parsers.vendor_parser import VendorParser


class AgilentGen5ImageParser(VendorParser[Data, Model]):
    DISPLAY_NAME = "Agilent Gen5 Image"
    RELEASE_STATE = ReleaseState.RECOMMENDED
    SUPPORTED_EXTENSIONS = AgilentGen5Reader.SUPPORTED_EXTENSIONS
    SCHEMA_MAPPER = Mapper

    def create_data(self, named_file_contents: NamedFileContents) -> Data:
        reader = AgilentGen5Reader(named_file_contents)

        header_data = HeaderData.create(
            reader.header_data, named_file_contents.original_file_path
        )
        read_data = ReadData.create(reader.sections["Procedure Details"])

        return Data(
            metadata=create_metadata(header_data),
            measurement_groups=create_results(
                get_results_section(reader),
                header_data,
                read_data,
                get_identifiers(reader.sections.get("Layout")),
            ),
        )



from dataclasses import dataclass

import pandas as pd

from allotropy.parsers.appbio_quantstudio_designandanalysis.appbio_quantstudio_designandanalysis_reader import (
    DesignQuantstudioReader,
)
from allotropy.parsers.appbio_quantstudio_designandanalysis.structure.generic.structure import (
    Result,
    Well,
    WellItem,
    WellList,
)
from allotropy.parsers.utils.pandas import SeriesData


class PresenceAbsenceResult(Result):
    @classmethod
    def get_genotyping_determination_result(cls, target_data: SeriesData) -> str | None:
        return target_data.get(str, "Call")

    @classmethod
    def get_genotyping_determination_method_setting(
        cls, target_data: SeriesData
    ) -> float | None:
        return target_data.get(float, "Threshold")


class PresenceAbsenceWellItem(WellItem):
    @classmethod
    def get_result_class(cls) -> type[Result]:
        return PresenceAbsenceResult


class PresenceAbsenceWell(Well):
    @classmethod
    def get_well_item_class(cls) -> type[WellItem]:
        return PresenceAbsenceWellItem


@dataclass(frozen=True)
class PresenceAbsenceWellList(WellList):
    @classmethod
    def get_well_class(cls) -> type[Well]:
        return PresenceAbsenceWell

    @classmethod
    def get_well_result_data(cls, reader: DesignQuantstudioReader) -> pd.DataFrame:
        return cls._add_data(
            data=reader.get_non_empty_sheet(cls.get_data_sheet()),
            extra_data=reader.get_non_empty_sheet("Target Call"),
            columns=[
                "Call",
            ],
        )



import pandas as pd

from allotropy.named_file_contents import NamedFileContents
from allotropy.parsers.lines_reader import determine_encoding
from allotropy.parsers.utils.pandas import read_csv


class ViCellBluReader:
    SUPPORTED_EXTENSIONS = "csv"

    @classmethod
    def read(cls, named_file_contents: NamedFileContents) -> pd.DataFrame:
        contents = named_file_contents.contents.read()
        encoding = (
            determine_encoding(contents, named_file_contents.encoding)
            if isinstance(contents, bytes)
            else None
        )
        named_file_contents.contents.seek(0)
        return read_csv(
            named_file_contents.contents, index_col=False, encoding=encoding
        )



from __future__ import annotations

from collections import defaultdict
from collections.abc import Iterator, Mapping
from contextlib import contextmanager
import json
from pathlib import Path, PurePosixPath, PureWindowsPath
import shutil
import tempfile
from typing import Any
from unittest import mock

from deepdiff import DeepDiff
from deepdiff.model import DiffLevel
from deepdiff.operator import BaseOperator
import numpy as np

from allotropy.allotrope.converter import structure
from allotropy.constants import ASM_CONVERTER_VERSION, DEFAULT_ENCODING
from allotropy.parser_factory import Vendor
from allotropy.to_allotrope import allotrope_from_file

DictType = Mapping[str, Any]


def _replace_asm_converter_version(allotrope_dict: DictType) -> DictType:
    new_dict = dict(allotrope_dict)
    for key, value in new_dict.items():
        if key == "data system document":
            value["ASM converter version"] = ASM_CONVERTER_VERSION
        if isinstance(value, dict):
            _replace_asm_converter_version(value)

    return new_dict


# List of keys in ASM with "identifier" that do not need to be unique.
# Currently, only "measurement identifier" and "calculated data document" identifier should be unique.
# However, it is better to have positive exceptions, so we don't accidentally miss a newly added unique identifier.
NON_UNIQUE_IDENTIFIERS = {
    "analytical method identifier",
    "assay bead identifier",
    "batch identifier",
    "data source identifier",
    "device identifier",
    "experimental data identifier",
    "flow cell identifier",
    "group identifier",
    "ifc identifier",
    "ligand identifier",
    "injection identifier",
    "location identifier",
    "source location identifier",
    "destination location identifier",
    "measurement method identifier",
    "sample identifier",
    "sensor chip identifier",
    "well location identifier",
    "source well location identifier",
    "destination well location identifier",
    "well plate identifier",
    "source well plate identifier",
    "destination well plate identifier",
    "well identifier",
    "assay identifier",
    "container identifier",
}


PATH_KEYS = {
    "POSIX path",
    "UNC path",
}


class PathComparison(BaseOperator):  # type: ignore[misc]
    # give_up_diffing stops diffing if returning True, here we use it to do a "real" comparison on path-like
    # leaf values. Because we only return True if we deem the paths equal, it's not really "giving up".
    def give_up_diffing(
        self, level: DiffLevel, diff_instance: DeepDiff  # noqa: ARG002
    ) -> bool:
        paths = [
            PureWindowsPath(raw_path) if "\\" in raw_path else PurePosixPath(raw_path)
            for raw_path in (level.t1, level.t2)
        ]
        return set(paths[0].parts) == set(paths[1].parts)


DEEPDIFF_PATH_COMPARATOR = PathComparison([f"{path_key}']$" for path_key in PATH_KEYS])


def _get_all_identifiers(asm: DictType) -> dict[str, list[str]]:
    all_identifiers = defaultdict(list)
    for key, value in asm.items():
        if isinstance(value, dict):
            for subkey, subvalue in _get_all_identifiers(value).items():
                all_identifiers[subkey].extend(subvalue)
        elif isinstance(value, list):
            for v in value:
                if isinstance(v, dict):
                    for subkey, subvalue in _get_all_identifiers(v).items():
                        all_identifiers[subkey].extend(subvalue)
        elif "identifier" in key:
            all_identifiers[key].append(value)
    return {key: list(value) for key, value in all_identifiers.items()}


def _get_all_with_key(target_key: str, item: Any) -> list[Any]:
    values = []
    if isinstance(item, dict):
        for key, value in item.items():
            if key == target_key:
                values.append(value)
            else:
                values.extend(_get_all_with_key(target_key, value))
    elif isinstance(item, list):
        for value in item:
            values.extend(_get_all_with_key(target_key, value))
    return values


def _validate_identifiers(asm: DictType) -> None:
    identifiers = _get_all_identifiers(asm)

    # Validate that all identifiers are unique.
    for key, value in identifiers.items():
        if key in NON_UNIQUE_IDENTIFIERS:
            continue
        if len(value) != len(set(value)):
            non_unique = [id_ for id_ in set(value) if value.count(id_) > 1]
            msg = f"Detected non-unique identifiers for key '{key}'. If this key should allow repeat values, add to NON_UNIQUE_IDENTIFIERS. Identifiers: {non_unique}"
            raise AssertionError(msg)

    # Validate that data source identifiers have a valid reference.
    data_source_ids = set(_get_all_with_key("data source identifier", asm))
    for ids in identifiers.values():
        data_source_ids -= set(ids)
    if data_source_ids:
        msg = f"data source identifiers {data_source_ids} do not have a valid identifier reference in the document."
        raise AssertionError(msg)


def _assert_allotrope_dicts_equal(
    expected: DictType,
    actual: DictType,
) -> None:
    expected_replaced = _replace_asm_converter_version(expected)
    ddiff = DeepDiff(
        expected_replaced,
        actual,
        ignore_type_in_groups=[(float, np.float64)],
        ignore_nan_inequality=True,
        custom_operators=[DEEPDIFF_PATH_COMPARATOR],
    )
    if ddiff:
        msg = f"allotropy output != expected: \n{ddiff.pretty()}"
        raise AssertionError(msg)


class TestIdGenerator:
    next_id: int
    prefix: str | None

    def __init__(self, prefix: str | None) -> None:
        self.prefix = f"{prefix}_" if prefix else ""
        self.next_id = 0

    def generate_id(self) -> str:
        current_id = f"{self.prefix}TEST_ID_{self.next_id}"
        self.next_id += 1
        return current_id


@contextmanager
def mock_uuid_generation(prefix: str | None = None) -> Iterator[None]:
    with mock.patch(
        "allotropy.parsers.utils.uuids._IdGeneratorFactory.get_id_generator",
        return_value=TestIdGenerator(prefix),
    ):
        yield


ROOT_DIR = Path(__file__).parent.parent.parent.parent


def get_testdata_dir(test_filepath: str) -> Path:
    return Path(Path(test_filepath).parent.relative_to(ROOT_DIR), "testdata")


def from_file(
    test_file: Path | str, vendor: Vendor, encoding: str | None = None
) -> DictType:
    with mock_uuid_generation(vendor.name):
        return allotrope_from_file(str(test_file), vendor, encoding=encoding)


def _write_actual_to_expected(
    allotrope_dict: DictType, expected_file: Path | str
) -> None:
    with tempfile.NamedTemporaryFile(mode="w+", encoding="UTF-8", delete=False) as tmp:
        json.dump(allotrope_dict, tmp, indent=4, ensure_ascii=False)
        tmp.write("\n")
        tmp.seek(0)
        # Get path to temp file using Pathlib to ensure Windows symbolic link compatibility.
        tmp_path = Path(tmp.name)
        # Ensure this file can be opened as JSON before we copy it
        with tmp_path.open() as tmp_file:
            json.load(tmp_file)
        shutil.copy(tmp_path, expected_file)


def validate_contents(
    allotrope_dict: DictType,
    expected_file: Path | str,
    write_actual_to_expected_on_fail: bool = False,  # noqa: FBT001, FBT002
) -> None:
    """Use the newly created allotrope_dict to validate the contents inside expected_file."""
    # Ensure that allotrope_dict can be written via json.dump()
    with tempfile.TemporaryFile(mode="w+", encoding=DEFAULT_ENCODING) as tmp:
        json.dump(allotrope_dict, tmp, ensure_ascii=False)

    try:
        with open(expected_file, encoding=DEFAULT_ENCODING) as f:
            expected_dict = json.load(f)
        _assert_allotrope_dicts_equal(expected_dict, allotrope_dict)
    except Exception as e:
        if write_actual_to_expected_on_fail:
            _write_actual_to_expected(allotrope_dict, expected_file)
            if isinstance(e, FileNotFoundError):
                msg = f"Missing expected output file '{expected_file}', writing expected output because 'write_actual_to_expected_on_fail=True'"
                raise AssertionError(msg) from e
            if isinstance(e, AssertionError) and "allotropy output != expected:" in str(
                e
            ):
                msg = f"Mismatch between actual and expected for '{expected_file}', writing expected output because 'write_actual_to_expected_on_fail=True'\n\n{e}"
                raise AssertionError(msg) from e
        raise

    # Ensure that allotrope_dict can be structured back into a python model.
    structure(allotrope_dict)

    # Ensure that all IDs are unique and that data source identifiers are valid references.
    _validate_identifiers(allotrope_dict)



import numpy as np
import pandas as pd

from allotropy.exceptions import AllotropeConversionError
from allotropy.named_file_contents import NamedFileContents
from allotropy.parsers.lines_reader import CsvReader, read_to_lines
from allotropy.parsers.thermo_fisher_visionlite.constants import (
    HEADER_COLS,
    SAMPLE_NAME_COLS,
)
from allotropy.parsers.utils.pandas import SeriesData
from allotropy.parsers.utils.values import assert_not_none


class ThermoFisherVisionliteReader:
    SUPPORTED_EXTENSIONS = "csv"
    header: SeriesData | None = None
    data: pd.DataFrame

    def __init__(self, named_file_contents: NamedFileContents) -> None:
        reader = CsvReader(read_to_lines(named_file_contents))

        if (first_line := reader.get()) is None:
            msg = "Unable to get data, empty file."
            raise AllotropeConversionError(msg)

        expected_required_columns = ("sample name", "well position")
        # If the first line in the file does not contain one of the required columns for QUANT or FIXED
        # experiments, then we expect the experiment type to be SCAN or KINETIC, and the first line to be a
        # header with comma separated key value pairs.
        if all(
            col_name not in first_line.lower() for col_name in expected_required_columns
        ):
            self.header = SeriesData(
                pd.Series(first_line.split(",")[:4], index=HEADER_COLS)
            )
            reader.pop()

        sep = "\t" if "\t" in first_line else ","
        data = assert_not_none(
            reader.pop_csv_block_as_df(
                header="infer", sep=sep, dtype={col: str for col in SAMPLE_NAME_COLS}
            )
        )
        self.data = data.replace(np.nan, None)



from allotropy.exceptions import AllotropeConversionError

NOT_APPLICABLE = "N/A"
# Used to fill an error value in a required ASM field.
NEGATIVE_ZERO = -0.0
# Used to fill in timestamp for required ASM fields when not available in source
DEFAULT_EPOCH_TIMESTAMP = "1970-01-01"

# Used to round well count to a reasonable number
POSSIBLE_WELL_COUNTS = [1, 2, 4, 6, 8, 12, 24, 48, 72, 96, 384, 1536, 3456]


def round_to_nearest_well_count(well_count: int) -> int | None:
    for possible_count in POSSIBLE_WELL_COUNTS:
        if well_count > possible_count:
            continue
        return possible_count
    return None


def get_well_count_by_well_ids(
    well_identifiers: list[int] | None = None, well_locations: list[str] | None = None
) -> int | None:
    if not well_identifiers and not well_locations:
        msg = "Must provide either well_identifiers or well_locations when determining plate size."
        raise AllotropeConversionError(msg)
    # Get well numbers via Well ID (1, 2, 3, ...) and well location (A1, B1, ...)
    well_number_by_ids = sorted(well_identifiers)[-1] if well_identifiers else 0
    well_number_by_position = 0
    if well_locations:
        largest_column = sorted([str(loc[0]) for loc in well_locations])[-1]
        largest_row = sorted(int(loc[1:]) for loc in well_locations)[-1]
        well_number_by_position = (
            ord(largest_column.upper()) - ord("A") + 1
        ) * largest_row
    largest_well_number = max(well_number_by_ids, well_number_by_position)
    if largest_well_number:
        return round_to_nearest_well_count(largest_well_number)
    return None
